#define  thd_t  (threadIdx.x % _NBT)
#define  thd_z  (threadIdx.x /(_NBT) % _NBZ)
#define  thd_y  (threadIdx.x /(_NBT*_NBZ) % _NBY)
#define  thd_x  (threadIdx.x /(_NBT*_NBZ*_NBY) % _NBX)
#define  blk_t  (blockIdx.x % _NGT)
#define  blk_z  (blockIdx.x /(_NGT) % _NGZ)
#define  blk_y  (blockIdx.x /(_NGT*_NGZ) % _NGY)
#define  blk_x  (blockIdx.x /(_NGT*_NGZ*_NGY) % _NGX)

static __global__ void 
cureverse_eo_kernel(int *ieo,int *ioe){
  *ioe = (*ieo+1)%2;
}

static __global__ void 
cuqclear_kernel(cuhqfield *y){
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].x = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].y = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].z = 0.0f;
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x].w = 0.0f;
  }
}

static __global__ void 
cuqmult1r_kernel(float *rcoef,cuhqfield *y, int *ieo){
// y = rcoef * y;
  __align__(16) float4 yri;
#pragma unroll 6
  for(unsigned int i=0;i<CSH;i++) {
    yri = (*y).b[blockIdx.x].yri[i].c[threadIdx.x];
    yri.x = yri.x * (*rcoef);
    yri.y = yri.y * (*rcoef);
    yri.z = yri.z * (*rcoef);
    yri.w = yri.w * (*rcoef);
    (*y).b[blockIdx.x].yri[i].c[threadIdx.x] = yri;
  }
}

static __global__ void 
cuqmult2r_accum_add_kernel(float *rcoef,cuhqfield *y, cuhqfield *w, int *ieo){
// w = rcoef * w + y;
  __align__(16) float4 yri,wri;
  yri = (*y).b[blockIdx.x].yri[0].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[0].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[0].c[threadIdx.x] = yri;
  yri = (*y).b[blockIdx.x].yri[1].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[1].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[1].c[threadIdx.x] = yri;
  yri = (*y).b[blockIdx.x].yri[2].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[2].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[2].c[threadIdx.x] = yri;
  yri = (*y).b[blockIdx.x].yri[3].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[3].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[3].c[threadIdx.x] = yri;
  yri = (*y).b[blockIdx.x].yri[4].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[4].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[4].c[threadIdx.x] = yri;
  yri = (*y).b[blockIdx.x].yri[5].c[threadIdx.x];
  wri = (*w).b[blockIdx.x].yri[5].c[threadIdx.x];
  yri.x += wri.x * (*rcoef);
  yri.y += wri.y * (*rcoef);
  yri.z += wri.z * (*rcoef);
  yri.w += wri.w * (*rcoef);
  (*w).b[blockIdx.x].yri[5].c[threadIdx.x] = yri;
}

static __global__ void 
cuqmult2c_accum_add_kernel(cuComplex *ccoef,cuhqfield *y, cuhqfield *w, int *ieo){
// w = ccoef * w + y;
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr = (*ccoef).x;
  ci = (*ccoef).y;
  cj =-(*ccoef).y;

  yr = (*y).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  yr.x += cr * wr.x;
  yr.x += cj * wi.x;
  yr.y += cr * wr.y;
  yr.y += cj * wi.y;
  yr.z += cr * wr.z;
  yr.z += cj * wi.z;
  yr.w += cr * wr.w;
  yr.w += cj * wi.w;
  yi.x += cr * wi.x;
  yi.x += ci * wr.x;
  yi.y += cr * wi.y;
  yi.y += ci * wr.y;
  yi.z += cr * wi.z;
  yi.z += ci * wr.z;
  yi.w += cr * wi.w;
  yi.w += ci * wr.w;
  (*w).b[blockIdx.x].yri[0    ].c[threadIdx.x] = yr;
  (*w).b[blockIdx.x].yri[0+COL].c[threadIdx.x] = yi;
  yr = (*y).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  yr.x += cr * wr.x;
  yr.x += cj * wi.x;
  yr.y += cr * wr.y;
  yr.y += cj * wi.y;
  yr.z += cr * wr.z;
  yr.z += cj * wi.z;
  yr.w += cr * wr.w;
  yr.w += cj * wi.w;
  yi.x += cr * wi.x;
  yi.x += ci * wr.x;
  yi.y += cr * wi.y;
  yi.y += ci * wr.y;
  yi.z += cr * wi.z;
  yi.z += ci * wr.z;
  yi.w += cr * wi.w;
  yi.w += ci * wr.w;
  (*w).b[blockIdx.x].yri[1    ].c[threadIdx.x] = yr;
  (*w).b[blockIdx.x].yri[1+COL].c[threadIdx.x] = yi;
  yr = (*y).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  yr.x += cr * wr.x;
  yr.x += cj * wi.x;
  yr.y += cr * wr.y;
  yr.y += cj * wi.y;
  yr.z += cr * wr.z;
  yr.z += cj * wi.z;
  yr.w += cr * wr.w;
  yr.w += cj * wi.w;
  yi.x += cr * wi.x;
  yi.x += ci * wr.x;
  yi.y += cr * wi.y;
  yi.y += ci * wr.y;
  yi.z += cr * wi.z;
  yi.z += ci * wr.z;
  yi.w += cr * wi.w;
  yi.w += ci * wr.w;
  (*w).b[blockIdx.x].yri[2    ].c[threadIdx.x] = yr;
  (*w).b[blockIdx.x].yri[2+COL].c[threadIdx.x] = yi;
}

static __global__ void 
cuqmult1c_accum_add_kernel(cuComplex *ccoef,cuhqfield *y, cuhqfield *w, int *ieo){
// w = w + ccoef * y;
  __align__(16) float cr,ci,cj;
  __align__(16) float4 wr,wi,yr,yi;
  cr = (*ccoef).x;
  ci = (*ccoef).y;
  cj =-(*ccoef).y;

  yr = (*y).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  wr.x += cr * yr.x;
  wr.x += cj * yi.x;
  wr.y += cr * yr.y;
  wr.y += cj * yi.y;
  wr.z += cr * yr.z;
  wr.z += cj * yi.z;
  wr.w += cr * yr.w;
  wr.w += cj * yi.w;
  wi.x += cr * yi.x;
  wi.x += ci * yr.x;
  wi.y += cr * yi.y;
  wi.y += ci * yr.y;
  wi.z += cr * yi.z;
  wi.z += ci * yr.z;
  wi.w += cr * yi.w;
  wi.w += ci * yr.w;
  (*w).b[blockIdx.x].yri[0    ].c[threadIdx.x] = wr;
  (*w).b[blockIdx.x].yri[0+COL].c[threadIdx.x] = wi;
  yr = (*y).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  wr.x += cr * yr.x;
  wr.x += cj * yi.x;
  wr.y += cr * yr.y;
  wr.y += cj * yi.y;
  wr.z += cr * yr.z;
  wr.z += cj * yi.z;
  wr.w += cr * yr.w;
  wr.w += cj * yi.w;
  wi.x += cr * yi.x;
  wi.x += ci * yr.x;
  wi.y += cr * yi.y;
  wi.y += ci * yr.y;
  wi.z += cr * yi.z;
  wi.z += ci * yr.z;
  wi.w += cr * yi.w;
  wi.w += ci * yr.w;
  (*w).b[blockIdx.x].yri[1    ].c[threadIdx.x] = wr;
  (*w).b[blockIdx.x].yri[1+COL].c[threadIdx.x] = wi;
  yr = (*y).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  wr.x += cr * yr.x;
  wr.x += cj * yi.x;
  wr.y += cr * yr.y;
  wr.y += cj * yi.y;
  wr.z += cr * yr.z;
  wr.z += cj * yi.z;
  wr.w += cr * yr.w;
  wr.w += cj * yi.w;
  wi.x += cr * yi.x;
  wi.x += ci * yr.x;
  wi.y += cr * yi.y;
  wi.y += ci * yr.y;
  wi.z += cr * yi.z;
  wi.z += ci * yr.z;
  wi.w += cr * yi.w;
  wi.w += ci * yr.w;
  (*w).b[blockIdx.x].yri[2    ].c[threadIdx.x] = wr;
  (*w).b[blockIdx.x].yri[2+COL].c[threadIdx.x] = wi;

}

#ifdef _OLD_REDUCTION_
static __global__ void 
cuprod_kernel(cuhqfield *y, cuhqfield *w, cuComplex *gri, int *ieo){
/************************************************************************
*$Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
* inner product y^+ * w
*   y : input quark field 
*   w : input quark field 
*  gr : partial sum real part
*  gi : partial sum imag parg
************************************************************************/
  float ywr = 0.0f;
  float yxr = 0.0f;
  float ywi = 0.0f;
  float yxi = 0.0f;

  {

  __align__(16) float4 wr,wi,yr,yi;
  yr = (*y).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[0    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[0+COL].c[threadIdx.x];
  ywr += yr.x * wr.x;
  ywr += yr.y * wr.y;
  ywr += yr.z * wr.z;
  ywr += yr.w * wr.w;
  yxr += yi.x * wi.x;
  yxr += yi.y * wi.y;
  yxr += yi.z * wi.z;
  yxr += yi.w * wi.w;
  ywi += yr.x * wi.x;
  ywi += yr.y * wi.y;
  ywi += yr.z * wi.z;
  ywi += yr.w * wi.w;
  yxi += yi.x * wr.x;
  yxi += yi.y * wr.y;
  yxi += yi.z * wr.z;
  yxi += yi.w * wr.w;
  yr = (*y).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[1    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[1+COL].c[threadIdx.x];
  ywr += yr.x * wr.x;
  ywr += yr.y * wr.y;
  ywr += yr.z * wr.z;
  ywr += yr.w * wr.w;
  yxr += yi.x * wi.x;
  yxr += yi.y * wi.y;
  yxr += yi.z * wi.z;
  yxr += yi.w * wi.w;
  ywi += yr.x * wi.x;
  ywi += yr.y * wi.y;
  ywi += yr.z * wi.z;
  ywi += yr.w * wi.w;
  yxi += yi.x * wr.x;
  yxi += yi.y * wr.y;
  yxi += yi.z * wr.z;
  yxi += yi.w * wr.w;
  yr = (*y).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  yi = (*y).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  wr = (*w).b[blockIdx.x].yri[2    ].c[threadIdx.x];
  wi = (*w).b[blockIdx.x].yri[2+COL].c[threadIdx.x];
  ywr += yr.x * wr.x;
  ywr += yr.y * wr.y;
  ywr += yr.z * wr.z;
  ywr += yr.w * wr.w;
  yxr += yi.x * wi.x;
  yxr += yi.y * wi.y;
  yxr += yi.z * wi.z;
  yxr += yi.w * wi.w;
  ywi += yr.x * wi.x;
  ywi += yr.y * wi.y;
  ywi += yr.z * wi.z;
  ywi += yr.w * wi.w;
  yxi += yi.x * wr.x;
  yxi += yi.y * wr.y;
  yxi += yi.z * wr.z;
  yxi += yi.w * wr.w;
  }

  ywr = ywr + yxr;
  ywi = ywi - yxi;
#ifdef _SF
  {
  int itb    = thd_t + blk_t*_NBT;
  int iz     = thd_z + blk_z*_NBZ;
  int iy     = thd_y + blk_y*_NBY;
  int ix     = thd_x + blk_x*_NBX;
  int ieoxyz = ((*ieo)+(iz+1)+(iy+1)+(ix+1))%2;
  float flag = (float)(1-ieoxyz*((_NTH - itb)/_NTH));
  ywr = ywr * flag;
  ywi = ywi * flag;
  }
#endif


  {
    __shared__ __align__(16) float sywr[CEIL],sywi[CEIL];
#if THREADS != CEIL
    sywr[CEIL - THREADS + threadIdx.x] = 0.0;
    sywi[CEIL - THREADS + threadIdx.x] = 0.0;
#endif
    __syncthreads();

    sywr[threadIdx.x] = ywr;
    sywi[threadIdx.x] = ywi;

    for (unsigned int s = CEIL/2; s > 32 ; s >>= 1){
      __syncthreads();
      if (threadIdx.x < s) {
        sywr[threadIdx.x] += sywr[threadIdx.x^s];
        sywi[threadIdx.x] += sywi[threadIdx.x^s];
      }
    }
    if (threadIdx.x < 32) {
      sywr[threadIdx.x] += sywr[threadIdx.x^32];
      sywi[threadIdx.x] += sywi[threadIdx.x^32];
      sywr[threadIdx.x] += sywr[threadIdx.x^16];
      sywi[threadIdx.x] += sywi[threadIdx.x^16];
      sywr[threadIdx.x] += sywr[threadIdx.x^8 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^8 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^4 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^4 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^2 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^2 ];
      sywr[threadIdx.x] += sywr[threadIdx.x^1 ];
      sywi[threadIdx.x] += sywi[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) {
       gri[blockIdx.x].x = sywr[0];
       gri[blockIdx.x].y = sywi[0];
    }
  }
}
#endif

template <unsigned int BlockSize>
static __global__ void 
quarkprod_kernel(cuhqfield *x,cuhqfield *y,float2 *z)
/***********************************************************
 * $Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
 *  improved inner-product
 *  z = < x^+ * y >
 *  x, y: input quark field
 *  z   : output intermediate results
 **********************************************************/
{
  __shared__ __align__(4) float wr[BlockSize],wi[BlockSize];

  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(BlockSize) + tid;
  unsigned int gridSize = BlockSize*gridDim.x;
  unsigned int n = THREADS*BLOCKS*COL*SPIN;

  wr[tid] = 0.0f; wi[tid] = 0.0f;

  float *xr = &((*x).b[0].yri[  0].c[0].x);
  float *xi = &((*x).b[0].yri[COL].c[0].x);
  float *yr = &((*y).b[0].yri[  0].c[0].x);
  float *yi = &((*y).b[0].yri[COL].c[0].x);

  float ar,ai,br,bi;

  unsigned int j;
  while (i < n) {
    j =  i % (THREADS*COL*SPIN)
      + (i / (THREADS*COL*SPIN)) * (THREADS*COL*SPIN*2);

    ar = xr[j];
    ai = xi[j];
    br = yr[j];
    bi = yi[j];

    wr[tid] += ar*br;
    wr[tid] += ai*bi;
    wi[tid] += ar*bi;
    wi[tid] -= ai*br;

    i += gridSize;
  }

  __syncthreads();

  if (BlockSize >= 512) {
    if (tid < 256) {
      wr[tid] += wr[tid + 256];
      wi[tid] += wi[tid + 256];
    }
    __syncthreads();
  }
  if (BlockSize >= 256) {
    if (tid < 128) {
      wr[tid] += wr[tid + 128];
      wi[tid] += wi[tid + 128];
    } 
    __syncthreads();
  }
  if (BlockSize >= 128) {
    if (tid <  64) {
      wr[tid] += wr[tid +  64];
      wi[tid] += wi[tid +  64];
    }
    __syncthreads();
  }

  //-------------

  if (tid < 32) {
    volatile float *sr = wr;
    volatile float *si = wi;

    if (BlockSize >= 64) {
      sr[tid] += sr[tid + 32];
      si[tid] += si[tid + 32];
    }
    if (BlockSize >= 32) {
      sr[tid] += sr[tid + 16];
      si[tid] += si[tid + 16];
    }
    if (BlockSize >= 16) {
      sr[tid] += sr[tid +  8];
      si[tid] += si[tid +  8];
    }
    if (BlockSize >=  8) {
      sr[tid] += sr[tid +  4];
      si[tid] += si[tid +  4];
    }
    if (BlockSize >=  4) {
      sr[tid] += sr[tid +  2];
      si[tid] += si[tid +  2];
    }
    if (BlockSize >= 2) {
      sr[tid] += sr[tid +  1];
      si[tid] += si[tid +  1];
    }
  }

  // write result
  if (tid == 0) {
    z[blockIdx.x].x = wr[0];
    z[blockIdx.x].y = wi[0];
  }

}

#ifdef _OLD_REDUCTION_
static __global__ void 
cuabs2_kernel(cuhqfield *y, float *g, int *ieo){
/************************************************************************
*$Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
*   y : input quark field 
************************************************************************/
  float ywr = 0.0f;
  {
  __align__(16) float4 yri;
  yri = (*y).b[blockIdx.x].yri[0].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  yri = (*y).b[blockIdx.x].yri[1].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  yri = (*y).b[blockIdx.x].yri[2].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  yri = (*y).b[blockIdx.x].yri[3].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  yri = (*y).b[blockIdx.x].yri[4].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  yri = (*y).b[blockIdx.x].yri[5].c[threadIdx.x];
  ywr += yri.x * yri.x;
  ywr += yri.y * yri.y;
  ywr += yri.z * yri.z;
  ywr += yri.w * yri.w;
  }
#ifdef _SF
  {
  int itb    = thd_t + blk_t*_NBT;
  int iz     = thd_z + blk_z*_NBZ;
  int iy     = thd_y + blk_y*_NBY;
  int ix     = thd_x + blk_x*_NBX;
  int ieoxyz = ((*ieo)+(iz+1)+(iy+1)+(ix+1))%2;
  float flag = (float)(1-ieoxyz*((_NTH - itb)/_NTH));
  ywr = ywr * flag;
  }
#endif


  {
    __shared__ __align__(16) float a[CEIL];
#if THREADS != CEIL
    a[CEIL - THREADS + threadIdx.x] = 0.0;
#endif

    __syncthreads();

    a[threadIdx.x] = ywr;

    for (unsigned int s = CEIL/2; s > 32 ; s >>= 1){
      __syncthreads();
      if (threadIdx.x < s) 
        a[threadIdx.x] += a[threadIdx.x^s];
    }
    if (threadIdx.x < 32) {
      a[threadIdx.x] += a[threadIdx.x^32];
      a[threadIdx.x] += a[threadIdx.x^16];
      a[threadIdx.x] += a[threadIdx.x^8 ];
      a[threadIdx.x] += a[threadIdx.x^4 ];
      a[threadIdx.x] += a[threadIdx.x^2 ];
      a[threadIdx.x] += a[threadIdx.x^1 ];
    }
    if (threadIdx.x==0) g[blockIdx.x] = a[0];
  }
}
#endif

template <unsigned int BlockSize>
static __global__ void
absolute_kernel(cuhqfield *x,float *z)
/***********************************************************
 * $Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
 *  square of absolute value of x.
 *  z = |z|^2
 **********************************************************/
{
  __shared__ __align__(4) float w[BlockSize];

  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(BlockSize) + tid;
  unsigned int gridSize = BlockSize*gridDim.x;
  unsigned int n = THREADS*BLOCKS*COL*SPIN*2;

  w[tid] = 0.0f;
  float *xx;
  xx = &((*x).b[0].yri[0].c[0].x);

  float a;
  while (i < n) {
    a = xx[i];
    w[tid] += a*a;

    i += gridSize;
  }

  __syncthreads();

  if (BlockSize >= 512) {
    if (tid < 256) { w[tid] += w[tid + 256]; } __syncthreads();
  }
  if (BlockSize >= 256) {
    if (tid < 128) { w[tid] += w[tid + 128]; } __syncthreads();
  }
  if (BlockSize >= 128) {
    if (tid <  64) { w[tid] += w[tid +  64]; } __syncthreads();
  }

  //-------------

  if (tid < 32) {
    volatile float *s = w;
    if (BlockSize >= 64) { s[tid] += s[tid + 32]; }
    if (BlockSize >= 32) { s[tid] += s[tid + 16]; }
    if (BlockSize >= 16) { s[tid] += s[tid +  8]; }
    if (BlockSize >=  8) { s[tid] += s[tid +  4]; }
    if (BlockSize >=  4) { s[tid] += s[tid +  2]; }
    if (BlockSize >=  2) { s[tid] += s[tid +  1]; }
  }

  // write result
  if (tid == 0) {
    z[blockIdx.x] = w[0];
  }

}

static __global__ void cu_rreduce(float *g){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) float a[];

  a[ith] = g[i] + g[i+blockDim.x];

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    __syncthreads();
    if (ith < s) a[ith] += a[ith^s];
  }
  if (ith==0) g[blockIdx.x] = a[0];
}

template <unsigned int BlockSize>
static __global__ void
cu_rreduce_kernel(float *a, float *b, unsigned int n)
/***********************************************************
 * $Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
 *  reduction of real value
 *  a : input  ( n elements )
 *  b : output ( "gridDim.x" wlements )
 **********************************************************/
{
  __shared__ __align__(4) float w[BlockSize];

  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(BlockSize) + tid;
  unsigned int gridSize = BlockSize*gridDim.x;

  w[tid] = 0.0f;

  while (i < n) {
    w[tid] += a[i];
    i += gridSize;
  }
  __syncthreads();

  if (BlockSize >= 512) {
    if (tid < 256) { w[tid] += w[tid + 256]; } __syncthreads();
  }
  if (BlockSize >= 256) {
    if (tid < 128) { w[tid] += w[tid + 128]; } __syncthreads();
  }
  if (BlockSize >= 128) {
    if (tid <  64) { w[tid] += w[tid +  64]; } __syncthreads();
  }

  //------------------

  if (tid < 32) {
    volatile float *s = w;
    if (BlockSize >= 64) { s[tid] += s[tid + 32]; }
    if (BlockSize >= 32) { s[tid] += s[tid + 16]; }
    if (BlockSize >= 16) { s[tid] += s[tid +  8]; }
    if (BlockSize >=  8) { s[tid] += s[tid +  4]; }
    if (BlockSize >=  4) { s[tid] += s[tid +  2]; }
    if (BlockSize >=  2) { s[tid] += s[tid +  1]; }
  }

  // write result
  if (tid == 0) {
    b[blockIdx.x] = w[0];
  }

}

static __global__ void cu_creduce(cuComplex *gri){
  unsigned int ith = threadIdx.x;
  unsigned int i   = blockIdx.x*(blockDim.x*2) + threadIdx.x;
  extern __shared__ __align__(16) cuComplex ari[];

  ari[ith].x = gri[i].x + gri[i+blockDim.x].x;
  ari[ith].y = gri[i].y + gri[i+blockDim.x].y;

  for (unsigned int s = blockDim.x/2; s > 0; s >>= 1){
    __syncthreads();
    if (ith < s) {
      ari[ith].x += ari[ith^s].x;
      ari[ith].y += ari[ith^s].y;
    }
  }
  if (ith==0) gri[blockIdx.x] = ari[0];
}

template <unsigned int BlockSize>
static __global__ void
cu_creduce_kernel(float2 *a,float2 *b,unsigned int n)
/***********************************************************
 * $Id: cuWqfAlgebra.h,v 1.13 2009/06/05 02:05:34 ishikawa Exp hayakawa $
 *  reduction of complex value
 *  a : input  ( n elements )
 *  b : output ( "gridDim.x" elements )
 **********************************************************/
{
  __shared__ __align__(16) float wr[BlockSize],wi[BlockSize];

  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x*(BlockSize) + tid;
  unsigned int gridSize = BlockSize*gridDim.x;
  wr[tid] = 0.0f; wi[tid] = 0.0f;

  while (i < n) {
    wr[tid] += a[i].x;
    wi[tid] += a[i].y;
    i += gridSize;
  }
  __syncthreads();

  if (BlockSize >= 512) {
    if (tid < 256) {
      wr[tid] += wr[tid + 256];
      wi[tid] += wi[tid + 256];
    }
    __syncthreads();
  }
  if (BlockSize >= 256) {
    if (tid < 128) {
      wr[tid] += wr[tid + 128];
      wi[tid] += wi[tid + 128];
    }
    __syncthreads();
  }
  if (BlockSize >= 128) {
    if (tid <  64) {
      wr[tid] += wr[tid + 64];
      wi[tid] += wi[tid + 64];
    }
    __syncthreads();
  }

  //------------------

  if (tid < 32) {
    volatile float *sr = wr;
    volatile float *si = wi;

    if (BlockSize >= 64) {
      sr[tid] += sr[tid + 32];
      si[tid] += si[tid + 32];
    }
    if (BlockSize >= 32) {
      sr[tid] += sr[tid + 16];
      si[tid] += si[tid + 16];
    }
    if (BlockSize >= 16) {
      sr[tid] += sr[tid +  8];
      si[tid] += si[tid +  8];
    }
    if (BlockSize >=  8) {
      sr[tid] += sr[tid + 4];
      si[tid] += si[tid + 4];
    }
    if (BlockSize >=  4) {
      sr[tid] += sr[tid + 2];
      si[tid] += si[tid + 2];
    }
    if (BlockSize >=  2) {
      sr[tid] += sr[tid + 1];
      si[tid] += si[tid + 1];
    }
  }

  // write result
  if (tid == 0) {
    b[blockIdx.x].x = wr[0];
    b[blockIdx.x].y = wi[0];
  }

}

static __global__ void cu_rreduce_get(float *g, float *a){
  *a = g[0];
}

static __global__ void cu_creduce_get(cuComplex *gri, cuComplex *a){
  *a = gri[0];
}

extern "C"
static void
getBlkAndThd(unsigned int n,unsigned int &blocks,unsigned int &threads)
{
  blocks  = 1;
  if (n <= MAXTHREADS) {
    threads = 1;
    while (threads < n) {
      threads <<= 1;
    }
  } else {
    threads = MAXTHREADS;
    while (blocks < MAXBLOCKS && blocks*MAXTHREADS < n) {
      blocks <<= 1;
    }
    blocks = min(MAXBLOCKS,blocks);
  }
}

extern "C"
static void 
cureverse_eo(int *ieo,int *ioe){
  cureverse_eo_kernel<<<1,1>>>(ieo,ioe);
  CUT_CHECK_ERROR("cureverse_eo Kernel execution failed");
}

#ifdef _OLD_REDUCTION_
extern "C"
static void 
old_cuabs2(cuhqfield *y, float *a, int *ieo){
  float *gdat;
  CUDA_SAFE_CALL(cudaMalloc((void**)&gdat,sizeof(float)*BLOCKS));

  cuabs2_kernel<<<BLOCKS,THREADS>>>(y,gdat,ieo);
  CUT_CHECK_ERROR("cuabs2 Kernel execution failed");

#if THREADS == CEIL
  cu_rreduce<<<1,BLOCKS/2,sizeof(float)*BLOCKS/2>>>(gdat);
  CUT_CHECK_ERROR("cu_rreduce Kernel execution failed");

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUDA_SAFE_CALL(cudaMemcpy(a,gdat,sizeof(float),cudaMemcpyDeviceToDevice));

#else
  // host side reduction
  float hdat[BLOCKS],aa;
  CUDA_SAFE_CALL(cudaMemcpy(hdat,gdat,sizeof(float)*BLOCKS,cudaMemcpyDeviceToHost));

  aa = 0.0;
  for (unsigned int i=0;i < BLOCKS; i++) {
    aa += hdat[i];
  }
  CUDA_SAFE_CALL(cudaMemcpy(a,&aa,sizeof(float),cudaMemcpyHostToDevice));
#endif
  CUDA_SAFE_CALL(cudaFree(gdat));

}
#endif

extern "C"
static void
cuabs2(cuhqfield *x, float *z, int *ieo)
{
  float *a;
  unsigned int blocks,threads;
  unsigned int ndata;

  getBlkAndThd(BLOCKS*THREADS*COL*SPIN*2,blocks,threads);

  CUDA_SAFE_CALL(cudaMalloc((void**)&a,sizeof(float)*blocks));

  switch (threads)
    {
    case 512:
      absolute_kernel<512><<<blocks,threads>>>(x,a); break;
    case 256:
      absolute_kernel<256><<<blocks,threads>>>(x,a); break;
    case 128:
      absolute_kernel<128><<<blocks,threads>>>(x,a); break;
    case  64:
      absolute_kernel< 64><<<blocks,threads>>>(x,a); break;
    case  32:
      absolute_kernel< 32><<<blocks,threads>>>(x,a); break;
    case  16:
      absolute_kernel< 16><<<blocks,threads>>>(x,a); break;
    case   8:
      absolute_kernel<  8><<<blocks,threads>>>(x,a); break;
    case   4:
      absolute_kernel<  4><<<blocks,threads>>>(x,a); break;
    case   2:
      absolute_kernel<  2><<<blocks,threads>>>(x,a); break;
    case   1:
      absolute_kernel<  1><<<blocks,threads>>>(x,a); break;
    }
  CUT_CHECK_ERROR("absolute_kerneld execution failed");

  while (blocks > 1) {
    ndata = blocks;
    getBlkAndThd(blocks,blocks,threads);
    switch (threads)
      {
      case 512:
        cu_rreduce_kernel<512><<<blocks,threads>>>(a,a,ndata); break;
      case 256:
        cu_rreduce_kernel<256><<<blocks,threads>>>(a,a,ndata); break;
      case 128:
        cu_rreduce_kernel<128><<<blocks,threads>>>(a,a,ndata); break;
      case  64:
        cu_rreduce_kernel< 64><<<blocks,threads>>>(a,a,ndata); break;
      case  32:
        cu_rreduce_kernel< 32><<<blocks,threads>>>(a,a,ndata); break;
      case  16:
        cu_rreduce_kernel< 16><<<blocks,threads>>>(a,a,ndata); break;
      case   8:
        cu_rreduce_kernel<  8><<<blocks,threads>>>(a,a,ndata); break;
      case   4:
        cu_rreduce_kernel<  4><<<blocks,threads>>>(a,a,ndata); break;
      case   2:
        cu_rreduce_kernel<  2><<<blocks,threads>>>(a,a,ndata); break;
      case   1:
        cu_rreduce_kernel<  1><<<blocks,threads>>>(a,a,ndata); break;
      }
    CUT_CHECK_ERROR("cu_rreduce_kernel execution failed");
  }

  CUDA_SAFE_CALL(cudaMemcpy(z,a,sizeof(float),cudaMemcpyDeviceToDevice));

  CUDA_SAFE_CALL(cudaFree(a));

}

#ifdef _OLD_REDUCTION_
extern "C"
static void 
old_cuprod(cuhqfield *y, cuhqfield *w, cuComplex *yw, int *ieo){
  cuComplex *gri;
  CUDA_SAFE_CALL(cudaMalloc((void**)&gri,sizeof(cuComplex)*BLOCKS));

  cuprod_kernel<<<BLOCKS,THREADS>>>(y,w,gri,ieo);
  CUT_CHECK_ERROR("cuprod Kernel execution failed");

#if THREADS == CEIL

  cu_creduce<<<1,BLOCKS/2,sizeof(cuComplex)*BLOCKS/2>>>(gri);
  CUT_CHECK_ERROR("cu_creduce Kernel execution failed");

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUDA_SAFE_CALL(cudaMemcpy(yw,gri,sizeof(cuComplex),cudaMemcpyDeviceToDevice));

#else
  // host side reducion
  cuComplex hri[BLOCKS],pp;
  CUDA_SAFE_CALL(cudaMemcpy(hri,gri,sizeof(cuComplex)*BLOCKS,cudaMemcpyDeviceToHost));

  pp.x = 0.0;
  pp.y = 0.0;
  for (unsigned int i=0;i < BLOCKS; i++) {
    pp.x += hri[i].x;
    pp.y += hri[i].y;
  }
  CUDA_SAFE_CALL(cudaMemcpy(yw,&pp,sizeof(cuComplex),cudaMemcpyHostToDevice));
#endif

  CUDA_SAFE_CALL(cudaFree(gri));

}
#endif

extern "C"
static void
cuprod(cuhqfield *x,cuhqfield *y,float2 *z,int *ieo)
{
  float2 *a;
  unsigned int blocks,threads;
  unsigned int ndata;

  getBlkAndThd(BLOCKS*THREADS*COL*SPIN,blocks,threads);

  CUDA_SAFE_CALL(cudaMalloc((void**)&a,sizeof(float2)*blocks));

  switch (threads)
    {
    case 512:
      quarkprod_kernel<512><<<blocks,threads>>>(x,y,a); break;
    case 256:
      quarkprod_kernel<256><<<blocks,threads>>>(x,y,a); break;
    case 128:
      quarkprod_kernel<128><<<blocks,threads>>>(x,y,a); break;
    case  64:
      quarkprod_kernel< 64><<<blocks,threads>>>(x,y,a); break;
    case  32:
      quarkprod_kernel< 32><<<blocks,threads>>>(x,y,a); break;
    case  16:
      quarkprod_kernel< 16><<<blocks,threads>>>(x,y,a); break;
    case   8:
      quarkprod_kernel<  8><<<blocks,threads>>>(x,y,a); break;
    case   4:
      quarkprod_kernel<  4><<<blocks,threads>>>(x,y,a); break;
    case   2:
      quarkprod_kernel<  2><<<blocks,threads>>>(x,y,a); break;
    case   1:
      quarkprod_kernel<  1><<<blocks,threads>>>(x,y,a); break;
    }
  CUT_CHECK_ERROR("quarkprod kernel execution failed");

  while (blocks > 1) {
    ndata = blocks;
    getBlkAndThd(blocks,blocks,threads);
    switch (threads)
      {
      case 512:
        cu_creduce_kernel<512><<<blocks,threads>>>(a,a,ndata); break;
      case 256:
        cu_creduce_kernel<256><<<blocks,threads>>>(a,a,ndata); break;
      case 128:
        cu_creduce_kernel<128><<<blocks,threads>>>(a,a,ndata); break;
      case  64:
        cu_creduce_kernel< 64><<<blocks,threads>>>(a,a,ndata); break;
      case  32:
        cu_creduce_kernel< 32><<<blocks,threads>>>(a,a,ndata); break;
      case  16:
        cu_creduce_kernel< 16><<<blocks,threads>>>(a,a,ndata); break;
      case   8:
        cu_creduce_kernel<  8><<<blocks,threads>>>(a,a,ndata); break;
      case   4:
        cu_creduce_kernel<  4><<<blocks,threads>>>(a,a,ndata); break;
      case   2:
        cu_creduce_kernel<  2><<<blocks,threads>>>(a,a,ndata); break;
      case   1:
        cu_creduce_kernel<  1><<<blocks,threads>>>(a,a,ndata); break;
      }
    CUT_CHECK_ERROR("cu_creduce kernel execution failed");
  }

  CUDA_SAFE_CALL(cudaMemcpy(z,a,sizeof(float2),cudaMemcpyDeviceToDevice));

  CUDA_SAFE_CALL(cudaFree(a));

}

extern "C"
static void 
cuqcopy(cuhqfield *ysrc, cuhqfield *wdest){
  CUDA_SAFE_CALL(cudaMemcpy(wdest, ysrc, sizeof(cuhqfield), cudaMemcpyDeviceToDevice));
}

extern "C"
static void 
cuqmult1c_accum_add(cuComplex *ccoef, cuhqfield *ysrc, cuhqfield *wdest, int *ieo){
  cuqmult1c_accum_add_kernel<<<BLOCKS,THREADS>>>(ccoef,ysrc,wdest,ieo);
  CUT_CHECK_ERROR("cuqmult1c_accum_add Kernel execution failed");
}

extern "C"
static void 
cuqmult1r(float *rcoef, cuhqfield *ysrc, int *ieo){
  cuqmult1r_kernel<<<BLOCKS,THREADS>>>(rcoef,ysrc,ieo);
  CUT_CHECK_ERROR("cuqmult1r Kernel execution failed");
}

extern "C"
static void 
cuqmult2r_accum_add(float *rcoef, cuhqfield *ysrc, cuhqfield *wdest, int *ieo){
  cuqmult2r_accum_add_kernel<<<BLOCKS,THREADS>>>(rcoef,ysrc,wdest,ieo);
  CUT_CHECK_ERROR("cuqmult2r_accum_add Kernel execution failed");
}

extern "C"
static void 
cuqmult2c_accum_add(cuComplex *ccoef, cuhqfield *ysrc, cuhqfield *wdest, int *ieo){
  cuqmult2c_accum_add_kernel<<<BLOCKS,THREADS>>>(ccoef,ysrc,wdest,ieo);
  CUT_CHECK_ERROR("cuqmult2c_accum_add Kernel execution failed");
}

extern "C"
static void cuqclear(cuhqfield *y){
  cuqclear_kernel<<<BLOCKS,THREADS>>>(y);
  CUT_CHECK_ERROR("cuqclear Kernel execution failed");
}

#undef  thd_t
#undef  thd_z
#undef  thd_y
#undef  thd_x
#undef  blk_t
#undef  blk_z
#undef  blk_y
#undef  blk_x
