extern "C"
static void 
#ifdef _CLOVER
cumult_mb_pre(float *kappa2,cuhqfield *ye,cuhqfield *yde,cuhgvfield *ue,cuhgvfield *uo,int *ieo,
              cuhclvfield *fclinve,cuhclvfield *fclinvo){
#else
cumult_mb_pre(float *kappa2,cuhqfield *ye,cuhqfield *yde,cuhgvfield *ue,cuhgvfield *uo,int *ieo){
#endif
/******************************************************************************
* Multiply even/odd-site preconditioned Dirac operator : yd = D y,
* where D = 1 - kappa^2 M M
*
*  kappa2 : hopping parameter (= -kappa^2)
*      ye : even/odd site fermion (input)
*     yde : even/odd site fermion (output)
*      ue : even/odd site gauge link
*      uo : odd/even site gauge link
*     ieo : 0 or 1 (0:even,1:odd)
* fclinve : even/odd site inverse clover term
* fclinvo : odd/even site inverse clover term
*
******************************************************************************/
  cuhqfield *Myo;
  int *ioe;

  CUDA_SAFE_CALL(cudaMalloc((void**)&Myo,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioe,sizeof(float)));
  cureverse_eo(ieo,ioe); // ioe = (ieo+1) % 2

#ifdef _CLOVER
  // Myo = F^-1_oo Moe ye
  // yde = F^-1_ee Meo Myo = F^-1_ee Meo F^-1_oo Moe ye
  cumult_wd( ye,Myo,uo,ue,ioe,fclinvo);
  cumult_wd(Myo,yde,ue,uo,ieo,fclinve);
#else
  // Myo = Moe ye
  // yde = Meo Myo = Meo Moe ye
  cumult_wd( ye,Myo,uo,ue,ioe);
  cumult_wd(Myo,yde,ue,uo,ieo);
#endif

  // yde = yo - kappa^2 yde
  cuqmult2r_accum_add(kappa2,ye,yde,ieo);

  CUDA_SAFE_CALL(cudaFree(Myo));
  CUDA_SAFE_CALL(cudaFree(ioe));
}
