/**********************************************************
* param-size.h
**********************************************************/
#define NDIM 4
#define NX   (_NTX/_NDIMX)
#define NY   (_NTY/_NDIMY)
#define NZ   (_NTZ/_NDIMZ)
#define NT   _NTT
#define NX1 (NX+1)
#define NY1 (NY+1)
#define NZ1 (NZ+1)
#define NT1 (NT+1)
#define NTH (NT/2)

#define COL    3
#define CLCL   (COL*COL)
#define SPIN   4
#define CLSP  (COL*SPIN)
#define CSH   (COL*SPIN/2)
#define CLSPH ((CLSP/2+1)*(CLSP/4))

#define GFIELD ((COL+1)*COL *NT*NZ*NY*NX*NDIM)
#define QFIELD ((COL+1)*SPIN*NT*NZ*NY*NX)
#define CFIELD (CLSPH*NT*NZ*NY*NX*2)
#define NSITE  (NT*NZ*NY*NX)
#define HNSITE (NTH*NZ*NY*NX)

/************************************************************************
*      nodeid : node number on whith this program is running, should be 0..NPE-1.
*        ipeo : indicates that this node begins with even/odd physical site.
*   ipsite(i) : node number in i-th direction,  should be ipsize(i)=0..NDIMi,
*               where i = 1,2,3 means i = x,y,z
* nodeidup(i) : indicates next upward node number in i-th direction.
* nodeiddn(i) : indicates next downward node number in i-th direction.
************************************************************************/
extern int param_size_mp_nodeid_;
extern int param_size_mp_ipeo_;
extern int ipsite[NDIM-1];
extern int mult_iter;
extern double copygtime;
extern double copyftime;

#ifdef __CUDACC__
#define _ALIG_ __align__(16)
#else
#define _ALIG_
#endif

/**********************************************************
* define structures
**********************************************************/
typedef struct _ALIG_ {
  double re[CLSP];
  double im[CLSP];
} su3fermi;

typedef struct _ALIG_ {
  su3fermi q[HNSITE];
} hqfield;

typedef struct _ALIG_ {
  double re[CLCL];
  double im[CLCL];
} su3mat;

typedef struct _ALIG_ {
  su3mat g[HNSITE];
} hgsfield;

typedef struct _ALIG_ {
  hgsfield v[NDIM];
} hgvfield;

typedef struct _ALIG_ {
  double f1r[CLSPH];
  double f1i[CLSPH];
  double f2r[CLSPH];
  double f2i[CLSPH];
} clv_i;

typedef struct _ALIG_ {
  clv_i f[HNSITE];
} hclvfield;

///////////////////////////////////////////////
typedef struct _ALIG_ {
  float x;
  float y;
  float z;
  float w;
} myfloat4;

typedef struct _ALIG_ {
  myfloat4 c[THREADS];
} myfloat4_thread;

///////////////////////////////////////////////
typedef struct _ALIG_ {
  myfloat4_thread yri[CSH];
} fhqf_thread;

typedef struct _ALIG_ {
  fhqf_thread b[BLOCKS];
} fhqfield;

///////////////////////////////////////////////
typedef struct _ALIG_ {
  myfloat4_thread uri[COL];
} fhgsf_thread;

typedef struct _ALIG_ {
  fhgsf_thread v[NDIM];
} fhgvf_thread;

typedef struct _ALIG_ {
  fhgvf_thread b[BLOCKS];
} fhgvfield;

///////////////////////////////////////////////
typedef struct _ALIG_ {
  myfloat4_thread fri[CLSPH];
} fclv_thread;

typedef struct _ALIG_ {
  fclv_thread b[BLOCKS];
} fhclvfield;


///////////////////////////////////////////////
typedef struct _ALIG_ {
  int t,z,y,x;
} d4site;

typedef struct _ALIG_ {
  d4site thread;
  d4site block;
} bksite;

static __inline__ int site(int it,int iz,int iy,int ix) {
  return (  (it-1)
          + (iz-1)*(NTH)
          + (iy-1)*(NTH)*(NZ)
          + (ix-1)*(NTH)*(NZ)*(NY) );
}

static __inline__ bksite d4site2bksite(d4site isite) {
  bksite jsite;
  jsite.thread.t = isite.t % _NBT;
   jsite.block.t = isite.t / _NBT;
  jsite.thread.z = isite.z % _NBZ;
   jsite.block.z = isite.z / _NBZ;
  jsite.thread.y = isite.y % _NBY;
   jsite.block.y = isite.y / _NBY;
  jsite.thread.x = isite.x % _NBX;
   jsite.block.x = isite.x / _NBX;
  return(jsite);
}

static __inline__ d4site bksite2d4site(bksite isite) {
  d4site jsite;
  jsite.t = isite.thread.t + isite.block.t * _NBT;
  jsite.z = isite.thread.z + isite.block.z * _NBZ;
  jsite.y = isite.thread.y + isite.block.y * _NBY;
  jsite.x = isite.thread.x + isite.block.x * _NBX;
  return(jsite);
}

static __inline__ int bksite2index(bksite isite) {
  return (  isite.thread.t
          + (_NBT)*(isite.thread.z)
          + (_NBT)*(_NBZ)*(isite.thread.y)
          + (_NBT)*(_NBZ)*(_NBY)*(isite.thread.x)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(isite.block.t)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(isite.block.z)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(_NGZ)*(isite.block.y)
          + (_NBT)*(_NBZ)*(_NBY)*(_NBX)*(_NGT)*(_NGZ)*(_NGY)*(isite.block.x));
}

static __inline__ int bksite2thread(bksite isite) {
  return (  isite.thread.t
          + (_NBT)*(isite.thread.z)
          + (_NBT)*(_NBZ)*(isite.thread.y)
          + (_NBT)*(_NBZ)*(_NBY)*(isite.thread.x));
}

static __inline__ int bksite2block(bksite isite) {
  return ( (isite.block.t)
          +(_NGT)*(isite.block.z)
          +(_NGT)*(_NGZ)*(isite.block.y)
          +(_NGT)*(_NGZ)*(_NGY)*(isite.block.x));
}

static __inline__ bksite index2bksite(int isite) {
  bksite jsite;
  jsite.thread.t = isite % _NBT;
  jsite.thread.z = isite /(_NBT) % _NBZ;
  jsite.thread.y = isite /(_NBT*_NBZ) % _NBY;
  jsite.thread.x = isite /(_NBT*_NBZ*_NBY) % _NBX;
  jsite.block.t  = isite /(_NBT*_NBZ*_NBY*_NBX) % _NGT;
  jsite.block.z  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT) % _NGZ;
  jsite.block.y  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT*_NGZ) % _NGY;
  jsite.block.x  = isite /(_NBT*_NBZ*_NBY*_NBX*_NGT*_NGZ*_NGY) % _NGX;
  return(jsite);
}

/////////////////////////////////////////////////////////////////////////////////////////
// define result structure
/////////////////////////////////////////////////////////////////////////////////////////
typedef struct {
  double time;
  float  inTime;
  int    iter;
  int    outIter;
  double residual;
} results;

/*
typedef struct {
  float time;
  int   iter;
} fresults;
*/
