/**************************************************************************
*
* Chronological guess inverter using single precision on GPU
*
* $Id: s_inv_chrolog.h,v 1.5 2011/03/28 09:38:21 hayakawa Exp $
**************************************************************************/
extern "C" void 
__S_INV_CHROLOG(fhqfield *x, int *nmax, int *inow, int *ndim,
                double *kappa, fhqfield *ye, fhqfield *dye, int *ieo) {
//
// GPU device pointer
//
  int *ieod;
  cuComplex *ctmpd,*ctmp2d;
  float *kappa2d,*rtmpd;
  cuhqfield *dyed;
  //  cuhqfield **dxd,**xd;
  // The above was replaced by
#ifdef _CHRON_
  cuhqfield *dxd[_CHRON_+1],*xd[_CHRON_];
#else
  cuhqfield *dxd[2],*xd[1];
#endif
//
// CPU variables
//
  cuComplex ctmp;
  float rtmp;
  float kappa2 = (float)(-(*kappa)*(*kappa));
  int i,j,n,itop,nn;

  CUDA_SAFE_CALL(cudaThreadSynchronize());
#ifdef _DEBUG_
  cutStartTimer(solvtime);
  fflush(stdout);
#endif

  n = *ndim;
  nn = n + 1;

  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod   ,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ctmpd  ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ctmp2d ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&rtmpd  ,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&kappa2d,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&dyed   ,sizeof(cuhqfield)));

  //  CUDA_SAFE_CALL(cudaMalloc((void**)&dxd    ,sizeof(cuhqfield*)*nn));
  //  CUDA_SAFE_CALL(cudaMalloc((void**)&xd     ,sizeof(cuhqfield*)*n));
  for (i=0;i<nn;i++) CUDA_SAFE_CALL(cudaMalloc((void**)&dxd[i],sizeof(cuhqfield)));
  for (i=0;i<n;i++)  CUDA_SAFE_CALL(cudaMalloc((void**)&xd[i] ,sizeof(cuhqfield)));

#ifdef _DEBUG_
  cutStartTimer(ysendtime);
#endif
  CUDA_SAFE_CALL(cudaMemcpy(kappa2d,&kappa2,sizeof(float),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ieod,ieo,sizeof(int),cudaMemcpyHostToDevice));
  ysend_count += sizeof(float) + sizeof(int);
  for (i=0;i<n;i++) {
    CUDA_SAFE_CALL(cudaMemcpy(xd[i],&(x[i]),sizeof(cuhqfield),cudaMemcpyHostToDevice));
    ysend_count += sizeof(cuhqfield);
  }
  CUDA_SAFE_CALL(cudaMemcpy(dxd[nn-1],ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));
  ysend_count += sizeof(cuhqfield);
#ifdef _DEBUG_
  cutStopTimer(ysendtime);
#endif


  itop = (((*inow)-1-1+(*nmax)) % (*nmax)) + 1;
  for (i=0;i<n;i++) {
#ifdef _CLOVER
    cumult_mb_pre(kappa2d,xd[itop-1],dxd[i],ued,uod,ieod,fclinved,fclinvod);
#else
    cumult_mb_pre(kappa2d,xd[itop-1],dxd[i],ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE;
    itop = ((itop-1-1+(*nmax)) % (*nmax)) + 1;
  }

  cuComplex RR[nn*nn];
  for (i=0;i<nn*nn;i++) {RR[i].x = 0.0f;RR[i].y = 0.0f;}

  for (i=0;i<nn;i++) {
    for (j=0;j<i;j++) {
      cuprod(dxd[j],dxd[i],ctmpd,ieod);
      cuGetComplexScalar(ctmpd,&(RR[j+i*nn]));
      cuNegative(ctmpd,ctmp2d);
      cuqmult1c_accum_add(ctmp2d,dxd[j],dxd[i],ieod);
      flop_count += QPROD + QMULT1C_ACCUM_ADD;
    }
    cuabs2(dxd[i],rtmpd,ieod);
    cuGetRealScalar(rtmpd,&rtmp);
    rtmp = sqrtf(rtmp);
    RR[i+i*nn].x = rtmp;
    rtmp = 1.0f/rtmp;
    cuPutRealScalar(&rtmp,rtmpd);
    cuqmult1r(rtmpd,dxd[i],ieod);
    flop_count += QABS2 + QMULT1R;
  }

  for (i=n-1;i>=0;i--) {
    for (j=i+1;j<n;j++) {
      ctmp.x = RR[i+j*nn].x * RR[j+(nn-1)*nn].x
             - RR[i+j*nn].y * RR[j+(nn-1)*nn].y;
      ctmp.y = RR[i+j*nn].x * RR[j+(nn-1)*nn].y
             + RR[i+j*nn].y * RR[j+(nn-1)*nn].x;
      RR[i+(nn-1)*nn].x = RR[i+(nn-1)*nn].x - ctmp.x;
      RR[i+(nn-1)*nn].y = RR[i+(nn-1)*nn].y - ctmp.y;
    }
    rtmp = 1.0f/RR[i+i*nn].x;
    RR[i+(nn-1)*nn].x = RR[i+(nn-1)*nn].x * rtmp;
    RR[i+(nn-1)*nn].y = RR[i+(nn-1)*nn].y * rtmp;
  }

  cuqclear(dyed);
  itop = (((*inow)-1-1+(*nmax)) % (*nmax)) + 1;
  for (i=0;i<n;i++) {
    cuPutComplexScalar(&(RR[i+(nn-1)*nn]),ctmpd);
    cuqmult1c_accum_add(ctmpd,xd[itop-1],dyed,ieod);
    flop_count += QMULT1C_ACCUM_ADD;
    itop = ((itop-1-1+(*nmax)) % (*nmax)) + 1;
  }

#ifdef _DEBUG_
  cutStartTimer(yrecvtime);
#endif
  CUDA_SAFE_CALL(cudaMemcpy(dye,dyed,sizeof(cuhqfield),cudaMemcpyDeviceToHost));
  yrecv_count += sizeof(cuhqfield);
#ifdef _DEBUG_
  cutStopTimer(yrecvtime);
#endif

  CUDA_SAFE_CALL(cudaFree(ieod));
  CUDA_SAFE_CALL(cudaFree(ctmpd));
  CUDA_SAFE_CALL(cudaFree(ctmp2d));
  CUDA_SAFE_CALL(cudaFree(rtmpd));
  CUDA_SAFE_CALL(cudaFree(kappa2d));
  CUDA_SAFE_CALL(cudaFree(dyed));
  for (i=0;i<nn;i++) CUDA_SAFE_CALL(cudaFree(dxd[i]));
  for (i=0;i<n;i++)  CUDA_SAFE_CALL(cudaFree(xd[i]));
  // The following was commented out
  //CUDA_SAFE_CALL(cudaFree(dxd));
  //CUDA_SAFE_CALL(cudaFree(xd));

#ifdef _DEBUG_
  fflush(stdout);
  cutStopTimer(solvtime);
#endif

}
