#ifdef _OPENMP
#include <omp.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cutil.h>

#include "cconfig.h"

namespace multiple_gpus {
// define _DEBUG_ to print out the residual history.
//
#define _DEBUG_
//#undef _DEBUG_

// Number of gpus
int num_gpus = _USE_GPU_;

// Number of OpenMp threads
int num_omps;


#ifdef _MYBIND_

extern "C" 
void get_num_gpu_(int *ngpu){
  *ngpu = num_gpus;
}

extern "C" 
void bind_cpu_to_gpu_(int *ngpu){

  int num_gpu_dev;

  CUDA_SAFE_CALL(cudaGetDeviceCount(&num_gpu_dev));

  if ( num_gpus > num_gpu_dev ) {
    printf("Specified number of GPUs exceeds the available number\n");
    exit(-1); 
  }

  *ngpu = num_gpus;

#ifdef _OPENMP
  num_omps = omp_get_max_threads();
#endif

#ifdef _DEBUG_
  printf("Number of Host OpenMP Threads : %d\n",num_omps);
  printf("Number of CUDA Device         : %d\n",num_gpus);
#endif

  if (num_gpus < 1) {
    printf("There is no CUDA Device. Program Stops.\n");
    exit(-1);
  }

  if (num_gpus > num_omps) {
    printf("Number of GPUs exceeds the Number of Host OpenMP Threads.\n");
    exit(-1);
  }

//  bind gpu card <=> openmp thread
#pragma omp parallel for
  for (int igx=0;igx<num_gpus;igx++) {
#ifdef _OPENMP
    int omp_thread_num  = omp_get_thread_num();
#else
    int omp_thread_num = 0;
#endif
    int cuda_device_num = igx % num_gpus;
#ifdef _DEBUG_
    printf("OpenMP Thread = %d  Cuda Device = %d\n",omp_thread_num,cuda_device_num);
#endif
    if (num_gpus > 1) CUDA_SAFE_CALL(cudaSetDevice(cuda_device_num));
  }  
#ifdef _DEBUG_
  fflush(stdout);
#endif
}



extern "C" void 
unbind_cpu_to_gpu_(){
#pragma omp parallel for
  for (int igx=0;igx<num_gpus;igx++) {
    if (num_gpus > 1) CUDA_SAFE_CALL(cudaThreadExit());
  }
#ifdef _DEBUG_
  fflush(stdout);
#endif
}
#endif

}
