#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <math.h>

#ifndef __CUDACC__
typedef struct { float x,y;} float2;
typedef struct { float x,y,z,w;} float4;
#endif

#include "cconfig.h"
#include "gpu_config.h"
//#include "param-size.h"
#include "cuparam-size.h"

/*******************************************
* copy(dobule<=>single) and site reordering
*******************************************/
void conv_q_vec_d2s_(double complex *vec, cuhqfield *fye, int *ieo, double *norm){
  int ix,iy,iz,itb,itb0,ic,is,ieoxyz,indx,iblk,ithd;
  double rtmp;
  d4site isite;
  bksite jsite;

  *norm = 0.0;
  rtmp = 0.0;
#pragma omp parallel for private(ix,iy,iz,ieoxyz,itb,itb0,ic,is,indx) reduction(+:rtmp)
  for(ix=1;ix<=NX;ix++){
  for(iy=1;iy<=NY;iy++){
  for(iz=1;iz<=NZ;iz++){
    ieoxyz = (*ieo+ix+iy+iz)%2;
#ifdef _SF
    for(itb=1;itb<=NTH-ieoxyz;itb++){
#else
    for(itb=1-ieoxyz;itb<=NTH-ieoxyz;itb++){
#endif
      itb0 = itb+ieoxyz;
      for(is=0;is<SPIN;is++){
      for(ic=0;ic<COL ;ic++){
        indx = ic
             + COL*is
             + COL*SPIN*(itb0-1)
             + COL*SPIN*NTH*(iz-1)
             + COL*SPIN*NTH*NZ*(iy-1)
             + COL*SPIN*NTH*NZ*NY*(ix-1);
        rtmp = rtmp
              + creal(vec[indx])*creal(vec[indx])
              + cimag(vec[indx])*cimag(vec[indx]);
      }
      }
    }
  }
  }
  }
  *norm = rtmp;
  *norm = sqrt(*norm);
  rtmp = 1.0/(*norm);

#pragma omp parallel for private(ix,iy,iz,ieoxyz,itb,itb0,isite,jsite,indx,iblk,ithd)
  for(ix=1;ix<=NX;ix++) {
  for(iy=1;iy<=NY;iy++) {
  for(iz=1;iz<=NZ;iz++) {
    ieoxyz = (*ieo+ix+iy+iz)%2;
    for(itb=1-ieoxyz;itb<=NTH-ieoxyz;itb++) {
      itb0 = itb + ieoxyz;
      isite.x = ix-1;
      isite.y = iy-1;
      isite.z = iz-1;
      isite.t = itb0-1;
      jsite = d4site2bksite(isite);

      indx = COL*SPIN*(itb0-1)
           + COL*SPIN*NTH*(iz-1)
           + COL*SPIN*NTH*NZ*(iy-1)
           + COL*SPIN*NTH*NZ*NY*(ix-1);

      iblk = bksite2block(jsite);
      ithd = bksite2thread(jsite);
#ifdef _SF
      if (itb > 0) {
#endif
      (*fye).b[iblk].yri[0].c[ithd].x = (float)(creal(vec[indx+ 0])*rtmp);
      (*fye).b[iblk].yri[0].c[ithd].y = (float)(creal(vec[indx+ 1])*rtmp);
      (*fye).b[iblk].yri[0].c[ithd].z = (float)(creal(vec[indx+ 2])*rtmp);
      (*fye).b[iblk].yri[0].c[ithd].w = (float)(creal(vec[indx+ 3])*rtmp);
      (*fye).b[iblk].yri[1].c[ithd].x = (float)(creal(vec[indx+ 4])*rtmp);
      (*fye).b[iblk].yri[1].c[ithd].y = (float)(creal(vec[indx+ 5])*rtmp);
      (*fye).b[iblk].yri[1].c[ithd].z = (float)(creal(vec[indx+ 6])*rtmp);
      (*fye).b[iblk].yri[1].c[ithd].w = (float)(creal(vec[indx+ 7])*rtmp);
      (*fye).b[iblk].yri[2].c[ithd].x = (float)(creal(vec[indx+ 8])*rtmp);
      (*fye).b[iblk].yri[2].c[ithd].y = (float)(creal(vec[indx+ 9])*rtmp);
      (*fye).b[iblk].yri[2].c[ithd].z = (float)(creal(vec[indx+10])*rtmp);
      (*fye).b[iblk].yri[2].c[ithd].w = (float)(creal(vec[indx+11])*rtmp);

      (*fye).b[iblk].yri[3].c[ithd].x = (float)(cimag(vec[indx+ 0])*rtmp);
      (*fye).b[iblk].yri[3].c[ithd].y = (float)(cimag(vec[indx+ 1])*rtmp);
      (*fye).b[iblk].yri[3].c[ithd].z = (float)(cimag(vec[indx+ 2])*rtmp);
      (*fye).b[iblk].yri[3].c[ithd].w = (float)(cimag(vec[indx+ 3])*rtmp);
      (*fye).b[iblk].yri[4].c[ithd].x = (float)(cimag(vec[indx+ 4])*rtmp);
      (*fye).b[iblk].yri[4].c[ithd].y = (float)(cimag(vec[indx+ 5])*rtmp);
      (*fye).b[iblk].yri[4].c[ithd].z = (float)(cimag(vec[indx+ 6])*rtmp);
      (*fye).b[iblk].yri[4].c[ithd].w = (float)(cimag(vec[indx+ 7])*rtmp);
      (*fye).b[iblk].yri[5].c[ithd].x = (float)(cimag(vec[indx+ 8])*rtmp);
      (*fye).b[iblk].yri[5].c[ithd].y = (float)(cimag(vec[indx+ 9])*rtmp);
      (*fye).b[iblk].yri[5].c[ithd].z = (float)(cimag(vec[indx+10])*rtmp);
      (*fye).b[iblk].yri[5].c[ithd].w = (float)(cimag(vec[indx+11])*rtmp);
#ifdef _SF
      } else {
      (*fye).b[iblk].yri[0].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[0].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[0].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[0].c[ithd].w = 0.0f;
      (*fye).b[iblk].yri[1].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[1].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[1].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[1].c[ithd].w = 0.0f;
      (*fye).b[iblk].yri[2].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[2].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[2].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[2].c[ithd].w = 0.0f;

      (*fye).b[iblk].yri[3].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[3].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[3].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[3].c[ithd].w = 0.0f;
      (*fye).b[iblk].yri[4].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[4].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[4].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[4].c[ithd].w = 0.0f;
      (*fye).b[iblk].yri[5].c[ithd].x = 0.0f;
      (*fye).b[iblk].yri[5].c[ithd].y = 0.0f;
      (*fye).b[iblk].yri[5].c[ithd].z = 0.0f;
      (*fye).b[iblk].yri[5].c[ithd].w = 0.0f;
      }
#endif

    }   // itb
  } } } // iz,iy,ix
}

void conv_q_vec_s2d_(cuhqfield *fye, double complex *vec, int *ieo, double *norm){
  int ix,iy,iz,itb,itb0,ic,is,ieoxyz,iblk,ithd;
  d4site isite;
  bksite jsite;
  int indx;
  double rtmp = *norm;

#pragma omp parallel for private(ix,iy,iz,ieoxyz,itb,itb0,isite,jsite,indx,iblk,ithd)
  for(ix=1;ix<=NX;ix++) {
  for(iy=1;iy<=NY;iy++) {
  for(iz=1;iz<=NZ;iz++) {
    ieoxyz = (*ieo+ix+iy+iz)%2;
    for(itb=1-ieoxyz;itb<=NTH-ieoxyz;itb++) {
      itb0 = itb + ieoxyz;
      isite.x = ix-1;
      isite.y = iy-1;
      isite.z = iz-1;
      isite.t = itb0-1;
      jsite = d4site2bksite(isite);
      indx = COL*SPIN*(itb0-1)
           + COL*SPIN*NTH*(iz-1)
           + COL*SPIN*NTH*NZ*(iy-1)
           + COL*SPIN*NTH*NZ*NY*(ix-1);
      iblk = bksite2block(jsite);
      ithd = bksite2thread(jsite);
#ifdef _SF
      if (itb > 0) {
#endif
      vec[indx+ 0] = ((double)(*fye).b[iblk].yri[0    ].c[ithd].x)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[0+COL].c[ithd].x)*(rtmp) * I;
      vec[indx+ 1] = ((double)(*fye).b[iblk].yri[0    ].c[ithd].y)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[0+COL].c[ithd].y)*(rtmp) * I;
      vec[indx+ 2] = ((double)(*fye).b[iblk].yri[0    ].c[ithd].z)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[0+COL].c[ithd].z)*(rtmp) * I;
      vec[indx+ 3] = ((double)(*fye).b[iblk].yri[0    ].c[ithd].w)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[0+COL].c[ithd].w)*(rtmp) * I;
      vec[indx+ 4] = ((double)(*fye).b[iblk].yri[1    ].c[ithd].x)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[1+COL].c[ithd].x)*(rtmp) * I;
      vec[indx+ 5] = ((double)(*fye).b[iblk].yri[1    ].c[ithd].y)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[1+COL].c[ithd].y)*(rtmp) * I;
      vec[indx+ 6] = ((double)(*fye).b[iblk].yri[1    ].c[ithd].z)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[1+COL].c[ithd].z)*(rtmp) * I;
      vec[indx+ 7] = ((double)(*fye).b[iblk].yri[1    ].c[ithd].w)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[1+COL].c[ithd].w)*(rtmp) * I;
      vec[indx+ 8] = ((double)(*fye).b[iblk].yri[2    ].c[ithd].x)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[2+COL].c[ithd].x)*(rtmp) * I;
      vec[indx+ 9] = ((double)(*fye).b[iblk].yri[2    ].c[ithd].y)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[2+COL].c[ithd].y)*(rtmp) * I;
      vec[indx+10] = ((double)(*fye).b[iblk].yri[2    ].c[ithd].z)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[2+COL].c[ithd].z)*(rtmp) * I;
      vec[indx+11] = ((double)(*fye).b[iblk].yri[2    ].c[ithd].w)*(rtmp)
                   + ((double)(*fye).b[iblk].yri[2+COL].c[ithd].w)*(rtmp) * I;
#ifdef _SF
      } else {
        vec[indx+ 0] = 0.0 + 0.0*I;
        vec[indx+ 1] = 0.0 + 0.0*I;
        vec[indx+ 2] = 0.0 + 0.0*I;
        vec[indx+ 3] = 0.0 + 0.0*I;
        vec[indx+ 4] = 0.0 + 0.0*I;
        vec[indx+ 5] = 0.0 + 0.0*I;
        vec[indx+ 6] = 0.0 + 0.0*I;
        vec[indx+ 7] = 0.0 + 0.0*I;
        vec[indx+ 8] = 0.0 + 0.0*I;
        vec[indx+ 9] = 0.0 + 0.0*I;
        vec[indx+10] = 0.0 + 0.0*I;
        vec[indx+11] = 0.0 + 0.0*I;
      }
#endif
    }   // itb
  } } } // iz,iy,ix
}

void conv_u_(double complex *due, cuhgvfield *fue, int *ieo){
  int ix,iy,iz,itb,itb0,mu,ic,jc,ieoxyz,iblk,ithd;
  d4site isite;
  bksite jsite;
  int indx;

#pragma omp parallel for private(mu,ix,iy,iz,ieoxyz,itb,itb0,isite,jsite,indx,iblk,ithd)
  for(mu=0;mu<NDIM;mu++) {
  for(ix=1;ix<=NX;ix++) {
  for(iy=1;iy<=NY;iy++) {
  for(iz=1;iz<=NZ;iz++) {
    ieoxyz = (*ieo+ix+iy+iz)%2;
    for(itb=1-ieoxyz;itb<=NTH-ieoxyz;itb++) {
      itb0 = itb + ieoxyz;
      isite.x = ix-1;
      isite.y = iy-1;
      isite.z = iz-1;
      isite.t = itb0-1;
      jsite = d4site2bksite(isite);
      indx = COL*COL*(itb)
           + COL*COL*(NTH+1)*(iz)
           + COL*COL*(NTH+1)*(NZ+2)*(iy)
           + COL*COL*(NTH+1)*(NZ+2)*(NY+2)*(ix)
           + COL*COL*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)*mu;
      iblk = bksite2block(jsite);
      ithd = bksite2thread(jsite);
      (*fue).b[iblk].v[mu].uri[0].c[ithd].x = (float)creal(due[indx+ 0]);
      (*fue).b[iblk].v[mu].uri[0].c[ithd].y = (float)creal(due[indx+ 1]);
      (*fue).b[iblk].v[mu].uri[0].c[ithd].z = (float)creal(due[indx+ 2]);
      (*fue).b[iblk].v[mu].uri[0].c[ithd].w = (float)creal(due[indx+ 3]);
      (*fue).b[iblk].v[mu].uri[1].c[ithd].x = (float)creal(due[indx+ 4]);
      (*fue).b[iblk].v[mu].uri[1].c[ithd].y = (float)creal(due[indx+ 5]);

      (*fue).b[iblk].v[mu].uri[1].c[ithd].z = (float)cimag(due[indx+ 0]);
      (*fue).b[iblk].v[mu].uri[1].c[ithd].w = (float)cimag(due[indx+ 1]);
      (*fue).b[iblk].v[mu].uri[2].c[ithd].x = (float)cimag(due[indx+ 2]);
      (*fue).b[iblk].v[mu].uri[2].c[ithd].y = (float)cimag(due[indx+ 3]);
      (*fue).b[iblk].v[mu].uri[2].c[ithd].z = (float)cimag(due[indx+ 4]);
      (*fue).b[iblk].v[mu].uri[2].c[ithd].w = (float)cimag(due[indx+ 5]);

    }   // itb
  } } } } // iz,iy,ix,mu
}

void conv_clv_(double complex *dclve, cuhclvfield *fclve, int *ieo){
  int ix,iy,iz,itb,itb0,ics,jcs,ieoxyz,iblk,ithd;
  d4site isite;
  bksite jsite;
  int indxi,indxj;
  double complex f1[CLSPH];
  double complex f2[CLSPH];

#pragma omp parallel for private(ix,iy,iz,ieoxyz,itb,itb0,isite,jsite,indxi,indxj,iblk,ithd,f1,f2,ics,jcs)
  for(ix=1;ix<=NX;ix++) {
  for(iy=1;iy<=NY;iy++) {
  for(iz=1;iz<=NZ;iz++) {
    ieoxyz = (*ieo+ix+iy+iz)%2;
    for(itb=1-ieoxyz;itb<=NTH-ieoxyz;itb++) {
      itb0 = itb + ieoxyz;
      isite.x = ix-1;
      isite.y = iy-1;
      isite.z = iz-1;
      isite.t = itb0-1;
      jsite = d4site2bksite(isite);
      iblk = bksite2block(jsite);
      ithd = bksite2thread(jsite);

      for(ics=0;ics<CLSPH;ics++) {
        indxi = ics
              + CLSPH*0
              + CLSPH*2*(itb0-1)
              + CLSPH*2*(NTH)*(iz-1)
              + CLSPH*2*(NTH)*(NZ)*(iy-1)
              + CLSPH*2*(NTH)*(NZ)*(NY)*(ix-1);
        indxj = ics
              + CLSPH*1
              + CLSPH*2*(itb0-1)
              + CLSPH*2*(NTH)*(iz-1)
              + CLSPH*2*(NTH)*(NZ)*(iy-1)
              + CLSPH*2*(NTH)*(NZ)*(NY)*(ix-1);

        f1[ics] = dclve[indxi];
        f2[ics] = dclve[indxj];
//printf("#FCLV1:%2d%4d%4d%4d%4d%4d%24.15E%24.15E%12d\n",*ieo,ics+1,itb,iz,iy,ix,creal(f1[ics]),cimag(f1[ics]),indxi);
//printf("#FCLV2:%2d%4d%4d%4d%4d%4d%24.15E%24.15E%12d\n",*ieo,ics+1,itb,iz,iy,ix,creal(f2[ics]),cimag(f2[ics]),indxj);
      }

      (*fclve).b[iblk].fri[ 0].c[ithd].x = (float)creal(f1[0]);
      (*fclve).b[iblk].fri[ 0].c[ithd].y = (float)creal(f1[1]);
      (*fclve).b[iblk].fri[ 0].c[ithd].z = (float)creal(f1[2]);
      (*fclve).b[iblk].fri[ 0].c[ithd].w = (float)creal(f1[3]);
      (*fclve).b[iblk].fri[ 1].c[ithd].x = (float)creal(f1[4]);
      (*fclve).b[iblk].fri[ 1].c[ithd].y = (float)creal(f1[5]);
      (*fclve).b[iblk].fri[ 1].c[ithd].z = (float)creal(f1[6]);
      (*fclve).b[iblk].fri[ 1].c[ithd].w = (float)creal(f1[7]);
      (*fclve).b[iblk].fri[ 2].c[ithd].x = (float)creal(f1[8]);
      (*fclve).b[iblk].fri[ 2].c[ithd].y = (float)creal(f1[9]);
      (*fclve).b[iblk].fri[ 2].c[ithd].z = (float)creal(f1[10]);
      (*fclve).b[iblk].fri[ 2].c[ithd].w = (float)creal(f1[11]);
      (*fclve).b[iblk].fri[ 3].c[ithd].x = (float)creal(f1[12]);
      (*fclve).b[iblk].fri[ 3].c[ithd].y = (float)creal(f1[13]);
      (*fclve).b[iblk].fri[ 3].c[ithd].z = (float)creal(f1[14]);
      (*fclve).b[iblk].fri[ 3].c[ithd].w = (float)creal(f1[15]);
      (*fclve).b[iblk].fri[ 4].c[ithd].x = (float)creal(f1[16]);
      (*fclve).b[iblk].fri[ 4].c[ithd].y = (float)creal(f1[17]);
      (*fclve).b[iblk].fri[ 4].c[ithd].z = (float)creal(f1[18]);
      (*fclve).b[iblk].fri[ 4].c[ithd].w = (float)creal(f1[19]);
      (*fclve).b[iblk].fri[ 5].c[ithd].x = (float)creal(f1[20]);

      (*fclve).b[iblk].fri[ 5].c[ithd].y = 0.0f;
      (*fclve).b[iblk].fri[ 5].c[ithd].z = (float)cimag(f1[1]);
      (*fclve).b[iblk].fri[ 5].c[ithd].w = (float)cimag(f1[2]);
      (*fclve).b[iblk].fri[ 6].c[ithd].x = (float)cimag(f1[3]);
      (*fclve).b[iblk].fri[ 6].c[ithd].y = (float)cimag(f1[4]);
      (*fclve).b[iblk].fri[ 6].c[ithd].z = (float)cimag(f1[5]);
      (*fclve).b[iblk].fri[ 6].c[ithd].w = 0.0f;
      (*fclve).b[iblk].fri[ 7].c[ithd].x = (float)cimag(f1[7]);
      (*fclve).b[iblk].fri[ 7].c[ithd].y = (float)cimag(f1[8]);
      (*fclve).b[iblk].fri[ 7].c[ithd].z = (float)cimag(f1[9]);
      (*fclve).b[iblk].fri[ 7].c[ithd].w = (float)cimag(f1[10]);
      (*fclve).b[iblk].fri[ 8].c[ithd].x = 0.0f;
      (*fclve).b[iblk].fri[ 8].c[ithd].y = (float)cimag(f1[12]);
      (*fclve).b[iblk].fri[ 8].c[ithd].z = (float)cimag(f1[13]);
      (*fclve).b[iblk].fri[ 8].c[ithd].w = (float)cimag(f1[14]);
      (*fclve).b[iblk].fri[ 9].c[ithd].x = 0.0f;
      (*fclve).b[iblk].fri[ 9].c[ithd].y = (float)cimag(f1[16]);
      (*fclve).b[iblk].fri[ 9].c[ithd].z = (float)cimag(f1[17]);
      (*fclve).b[iblk].fri[ 9].c[ithd].w = 0.0f;
      (*fclve).b[iblk].fri[10].c[ithd].x = (float)cimag(f1[19]);
      (*fclve).b[iblk].fri[10].c[ithd].y = 0.0f;

      (*fclve).b[iblk].fri[10].c[ithd].z = (float)creal(f2[0]);
      (*fclve).b[iblk].fri[10].c[ithd].w = (float)creal(f2[1]);
      (*fclve).b[iblk].fri[11].c[ithd].x = (float)creal(f2[2]);
      (*fclve).b[iblk].fri[11].c[ithd].y = (float)creal(f2[3]);
      (*fclve).b[iblk].fri[11].c[ithd].z = (float)creal(f2[4]);
      (*fclve).b[iblk].fri[11].c[ithd].w = (float)creal(f2[5]);
      (*fclve).b[iblk].fri[12].c[ithd].x = (float)creal(f2[6]);
      (*fclve).b[iblk].fri[12].c[ithd].y = (float)creal(f2[7]);
      (*fclve).b[iblk].fri[12].c[ithd].z = (float)creal(f2[8]);
      (*fclve).b[iblk].fri[12].c[ithd].w = (float)creal(f2[9]);
      (*fclve).b[iblk].fri[13].c[ithd].x = (float)creal(f2[10]);
      (*fclve).b[iblk].fri[13].c[ithd].y = (float)creal(f2[11]);
      (*fclve).b[iblk].fri[13].c[ithd].z = (float)creal(f2[12]);
      (*fclve).b[iblk].fri[13].c[ithd].w = (float)creal(f2[13]);
      (*fclve).b[iblk].fri[14].c[ithd].x = (float)creal(f2[14]);
      (*fclve).b[iblk].fri[14].c[ithd].y = (float)creal(f2[15]);
      (*fclve).b[iblk].fri[14].c[ithd].z = (float)creal(f2[16]);
      (*fclve).b[iblk].fri[14].c[ithd].w = (float)creal(f2[17]);
      (*fclve).b[iblk].fri[15].c[ithd].x = (float)creal(f2[18]);
      (*fclve).b[iblk].fri[15].c[ithd].y = (float)creal(f2[19]);
      (*fclve).b[iblk].fri[15].c[ithd].z = (float)creal(f2[20]);

      (*fclve).b[iblk].fri[15].c[ithd].w = 0.0f;
      (*fclve).b[iblk].fri[16].c[ithd].x = (float)cimag(f2[1]);
      (*fclve).b[iblk].fri[16].c[ithd].y = (float)cimag(f2[2]);
      (*fclve).b[iblk].fri[16].c[ithd].z = (float)cimag(f2[3]);
      (*fclve).b[iblk].fri[16].c[ithd].w = (float)cimag(f2[4]);
      (*fclve).b[iblk].fri[17].c[ithd].x = (float)cimag(f2[5]);
      (*fclve).b[iblk].fri[17].c[ithd].y = 0.0f;
      (*fclve).b[iblk].fri[17].c[ithd].z = (float)cimag(f2[7]);
      (*fclve).b[iblk].fri[17].c[ithd].w = (float)cimag(f2[8]);
      (*fclve).b[iblk].fri[18].c[ithd].x = (float)cimag(f2[9]);
      (*fclve).b[iblk].fri[18].c[ithd].y = (float)cimag(f2[10]);
      (*fclve).b[iblk].fri[18].c[ithd].z = 0.0f;
      (*fclve).b[iblk].fri[18].c[ithd].w = (float)cimag(f2[12]);
      (*fclve).b[iblk].fri[19].c[ithd].x = (float)cimag(f2[13]);
      (*fclve).b[iblk].fri[19].c[ithd].y = (float)cimag(f2[14]);
      (*fclve).b[iblk].fri[19].c[ithd].z = 0.0f;
      (*fclve).b[iblk].fri[19].c[ithd].w = (float)cimag(f2[16]);
      (*fclve).b[iblk].fri[20].c[ithd].x = (float)cimag(f2[17]);
      (*fclve).b[iblk].fri[20].c[ithd].y = 0.0f;
      (*fclve).b[iblk].fri[20].c[ithd].z = (float)cimag(f2[19]);
      (*fclve).b[iblk].fri[20].c[ithd].w = 0.0f;

    } // itb
  } } } // iz,iy,ix
  fflush(stdout);
}
