#ifndef _MAIN_GPU_SOLVER_H
#define _MAIN_GPU_SOLVER_H
////////////////////////////////////////////////////////////////////////////////
// Single precision GPU solver
//
// $Id: main_gpu_solver.h,v 1.2 2011/10/23 14:15:55 ishikawa Exp ishikawa $
////////////////////////////////////////////////////////////////////////////////
#ifdef _DEBUG_
//
// Timer names
//
unsigned int gsendtime;
unsigned int ysendtime;
unsigned int yrecvtime;
unsigned int solvtime;
unsigned int multtime;
#endif

// flop counter
long long int flop_count;
long long int flop_mult_count;
long long int gsend_count;
long long int ysend_count;
long long int yrecv_count;

//
// Texture fetching for the link variables.
//
#ifdef _USE_TEXFETCH_
texture<float4,1,cudaReadModeElementType> tex_ueri;
texture<float4,1,cudaReadModeElementType> tex_uori;
texture<float4,1,cudaReadModeElementType> tex_yori;
#endif

#include "cuMyComplex.h"
#include "YMACRO.h"
#include "cuMatVec.h"
#include "cuWqfAlgebra.h"
#include "cu_mult_forw_t.h"
#include "cu_mult_forw_z.h"
#include "cu_mult_forw_y.h"
#include "cu_mult_forw_x.h"
#include "cu_mult_back_t.h"
#include "cu_mult_back_z.h"
#include "cu_mult_back_y.h"
#include "cu_mult_back_x.h"
#include "cu_mult_clover_term.h"
#include "assign_mult_hop.h"
#include "assign_mult_eoprec_wd.h"

//
// global device memory for LINK/CLOVER fields.
//
cuhgvfield *ued,*uod;
#ifdef _CLOVER
cuhclvfield *fclinved;
cuhclvfield *fclinvod;
#endif

extern "C" void 
#ifdef _CLOVER
__NEW_GPU_SOLVER(cuhgvfield *ue,cuhgvfield *uo,int *ieo,
                cuhclvfield *fclinve,cuhclvfield *fclinvo) {
#else
__NEW_GPU_SOLVER(cuhgvfield *ue,cuhgvfield *uo,int *ieo) {
#endif
//
// send LINK/CLOVER data from HOST to GPU
//

#ifndef _singlePU
  printf(" GPU solver does not support Parallel environment (MPI).\n");
  printf(" Program stops.\n");
  exit(-1);
#endif

#ifdef _DEBUG_
//
// Initialize timers.
//
  fflush(stdout);
  printf("# Start NEW GPU Solver: %s \n",__func__);
  printf("# sizeof(cuhqfield)  =%lld\n",sizeof(cuhqfield));
  printf("# sizeof(cuhgvfield) =%lld\n",sizeof(cuhgvfield));
  printf("# sizeof(cuhclvfield) =%lld\n",sizeof(cuhclvfield));
  fflush(stdout);
  cutCreateTimer(&gsendtime);
  cutCreateTimer(&ysendtime);
  cutCreateTimer(&yrecvtime);
  cutCreateTimer(&solvtime);
  cutCreateTimer(&multtime);
  cutResetTimer(gsendtime);
  cutResetTimer(ysendtime);
  cutResetTimer(yrecvtime);
  cutResetTimer(solvtime);
  cutResetTimer(multtime);
#endif

  gsend_count=0;
  ysend_count=0;
  yrecv_count=0;
  flop_count=0;
  flop_mult_count=0;

#ifdef _DEBUG_
  cutStartTimer(gsendtime);
#endif
  CUDA_SAFE_CALL(cudaMalloc((void**)&ued,sizeof(cuhgvfield)));
  CUT_CHECK_ERROR("cudaMalloc ued faild");

  CUDA_SAFE_CALL(cudaMalloc((void**)&uod,sizeof(cuhgvfield)));
  CUT_CHECK_ERROR("cudaMalloc uod faild");

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaMalloc((void**)&fclinved,sizeof(cuhclvfield)));
  CUT_CHECK_ERROR("cudaMalloc fclinved faild");

  CUDA_SAFE_CALL(cudaMalloc((void**)&fclinvod,sizeof(cuhclvfield)));
  CUT_CHECK_ERROR("cudaMalloc fclinvod faild");
#endif

  CUDA_SAFE_CALL(cudaMemcpy(ued,ue,sizeof(cuhgvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy ued faild");

  CUDA_SAFE_CALL(cudaMemcpy(uod,uo,sizeof(cuhgvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy uod faild");

  gsend_count = sizeof(cuhgvfield)*2;

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaMemcpy(fclinved,fclinve,sizeof(cuhclvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy fclinved faild");

  CUDA_SAFE_CALL(cudaMemcpy(fclinvod,fclinvo,sizeof(cuhclvfield),cudaMemcpyHostToDevice));
  CUT_CHECK_ERROR("cudaMemcpy fclinvod faild");

  gsend_count += sizeof(cuhclvfield)*2;
#endif

#ifdef _DEBUG_
  cutStopTimer(gsendtime);
#endif
}

extern "C" void 
__DELETE_GPU_SOLVER() {

#ifdef _DEBUG_
  fflush(stdout);
  printf("# Start DELETE GPU Solver: %s\n",__func__);
#endif
//
// Finalize GPU solver
//
  CUDA_SAFE_CALL(cudaFree(ued));
  CUT_CHECK_ERROR("cudaFree ued faild");

  CUDA_SAFE_CALL(cudaFree(uod));
  CUT_CHECK_ERROR("cudaFree uod faild");

#ifdef _CLOVER
  CUDA_SAFE_CALL(cudaFree(fclinved));
  CUT_CHECK_ERROR("cudaFree fclinved faild");

  CUDA_SAFE_CALL(cudaFree(fclinvod));
  CUT_CHECK_ERROR("cudaFree fclinvod faild");
#endif

#ifdef _DEBUG_
  double etime_solv  = ((double)cutGetTimerValue( solvtime))*1.0e-3;
  double etime_gsend = ((double)cutGetTimerValue(gsendtime))*1.0e-3;
  double etime_ysend = ((double)cutGetTimerValue(ysendtime))*1.0e-3;
  double etime_yrecv = ((double)cutGetTimerValue(yrecvtime))*1.0e-3;
  double etime_mult  = ((double)cutGetTimerValue( multtime))*1.0e-3;
  cutDeleteTimer(solvtime);
  cutDeleteTimer(gsendtime);
  cutDeleteTimer(ysendtime);
  cutDeleteTimer(yrecvtime);
  cutDeleteTimer(multtime);
  double gflops      = ((((double)HNSITE)*((double)flop_count     ))/etime_solv)*1.e-9;
  double gflops_mult = ((((double)HNSITE)*((double)flop_mult_count))/etime_mult)*1.e-9;
  long long int tflop_count      = HNSITE*flop_count;
  long long int tflop_mult_count = HNSITE*flop_mult_count;

  printf("%% Mult_GFLOPS_SP      : %15.7F\n",gflops_mult);
  printf("%% Mult_FLOP_SP        : %15.7lld\n",tflop_mult_count);
  printf("%% Mult_SP   ETIME(sec): %15.7F\n",etime_mult);
  printf("%% Solve_GFLOPS_SP     : %15.7F\n",gflops);
  printf("%% Solve_FLOP_SP       : %15.7lld\n",tflop_count);
  printf("%% Solve_SP  ETIME(sec): %15.7F\n",etime_solv);

  float gsendsize = (float)(gsend_count)/1024.0/1024.0;
  float ysendsize = (float)(ysend_count)/1024.0/1024.0;
  float yrecvsize = (float)(yrecv_count)/1024.0/1024.0;

  printf("%% GPU_GSend ETIME(sec): %15.7F\n",etime_gsend);
  printf("%% GPU_GSend  SIZE(MiB): %15.7F\n",gsendsize);
  printf("%% GPU_GSend  (MiB/sec): %15.7F\n",gsendsize/etime_gsend);
  printf("%% GPU_YSend ETIME(sec): %15.7F\n",etime_ysend);
  printf("%% GPU_YSend  SIZE(MiB): %15.7F\n",ysendsize);
  printf("%% GPU_YSend  (MiB/sec): %15.7F\n",ysendsize/etime_ysend);
  printf("%% GPU_YRecv ETIME(sec): %15.7F\n",etime_yrecv);
  printf("%% GPU_YRecv  SIZE(MiB): %15.7F\n",yrecvsize);
  printf("%% GPU_YRecv  (MiB/sec): %15.7F\n",yrecvsize/etime_yrecv);
  fflush(stdout);
#endif
}

extern "C" void 
__S_BICGSTAB_HMC_GPU(double *tol,int *iter, double *kappa, cuhqfield *ye, cuhqfield *yde, int *ieo) {
//
// GPU Solver
//

//===========================================================
// host variables
//===========================================================
  int ioe = (*ieo+1)%2;
  int niter = 5000;
  float mkappa2 = -(float)((*kappa)*(*kappa));
  float tol2 = (float)((*tol)*(*tol));
  float rtmp0;
  cuComplex alp,bet,omg,rho0,rho1,ctmp0,ctmp1;
#ifdef _DEBUG_
  float err;
#endif

//===========================================================
// device variables
//===========================================================
  int *ieod,*ioed;
  float *mkappa2d;

  cuComplex *alpd,*nalpd,*betd,*omgd,*nomgd;
  cuhqfield *yed,*yded,*red,*rted,*ped,*qed,*ted;

  CUDA_SAFE_CALL(cudaThreadSynchronize());
#ifdef _DEBUG_
  fflush(stdout);
  printf("# Start GPU Solver: %s\n",__func__);
  printf("# kappa=%24.15e\n",*kappa);
  printf("#   tol=%24.15e\n",*tol);
  printf("#  iter=%10d\n",*iter);
  printf("#   ieo=%2d\n",*ieo);
  cutStartTimer(solvtime);
  fflush(stdout);
#endif

//======================================
// allocate device memory
//======================================
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));

  CUDA_SAFE_CALL(cudaMalloc((void**) &alpd  ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&nalpd  ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&betd   ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**) &omgd  ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&nomgd  ,sizeof(cuComplex)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&mkappa2d,sizeof(float)));

  CUDA_SAFE_CALL(cudaMalloc((void**)&yed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&yded,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&red ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&rted,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ped ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&qed ,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ted ,sizeof(cuhqfield)));

//======================================
// Send data from host to device
//======================================
#ifdef _DEBUG_
  cutStartTimer(ysendtime);
#endif
  assign(*mkappa2d,mkappa2);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(yed,ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));
  ysend_count += sizeof(int)*2 + sizeof(float) + sizeof(cuhqfield);
#ifdef _DEBUG_
  cutStopTimer(ysendtime);
#endif

///////////////////////////////////////////////////////////////////////////////
  // set x_0 = 0 (yde=0)
  clear(yded);

  // r  = b ( = y : rhs of equation )
  // rt = r
  // p  = r
  assign( red,yed);
  assign(rted,red);
  assign( ped,red);

  // rho0 = |re|^2
  rtmp0 = abs2(red,ieod);         // rtmpd = |red|^2
  rho0.x = rtmp0;
  rho0.y = 0.0f;
  flop_count += QABS2;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
  printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {
    // q = A p
#ifdef _CLOVER
    assign_mult_eoprec_wd(mkappa2d,qed,ped,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(mkappa2d,qed,ped,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    // alp = rho0/<rt|q>
    ctmp0 = prod(rted,qed,ieod);
    alp = assign_div(rho0,ctmp0); // alp  = rho0/ctmp0
    assign(*alpd,alp);
    ctmp0 = assign_neg(alp);      // ctmp0 = -alp
    assign(*nalpd,ctmp0);
    flop_count += QPROD;

    // x = x + alp p
    // r = r - alp q
    accum_add_mult(yded, alpd,ped);
    accum_add_mult( red,nalpd,qed);
    flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red,ieod);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    flop_count += QABS2;
    if (rtmp0 < tol2) break;

    // t = D r
#ifdef _CLOVER
    assign_mult_eoprec_wd(mkappa2d,ted,red,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(mkappa2d,ted,red,ued,uod,ieod);
#endif
    flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    // omg = <t|r>/<t|t>
    rtmp0 = abs2(ted,ieod);
    ctmp0 = prod(ted,red,ieod);
    omg = assign_div(ctmp0,rtmp0);
    assign(*omgd,omg);
    ctmp0 = assign_neg(omg);
    assign(*nomgd,ctmp0);
    flop_count += QABS2 + QPROD;

    // x = x + omg r
    // r = r - omg t
    accum_add_mult(yded, omgd,red);
    accum_add_mult( red,nomgd,ted);
    flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red,ieod);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    flop_count += QABS2;
    if (rtmp0 < tol2) break;

    // rho1 = <rt|r>
    rho1 = prod(rted,red,ieod);
    flop_count += QPROD;

    // bet = (alp/omg)*(rho1/rho0)
    ctmp0  = assign_div(alp,omg);
    ctmp1 = assign_div(rho1,rho0);
    bet   = assign_mul(ctmp0,ctmp1);
    assign(*betd,bet);

    // rho0 = rho1;
    rho0 = rho1;

    // p = r + bet ( p - omg q)
    accum_add_mult(ped,nomgd,qed);
    accum_mult_add(ped, betd,red);
    flop_count += QMULT1C_ACCUM_ADD+QMULT2C_ACCUM_ADD;
  } // end while

#ifdef _DEBUG_
///////////////////////////////////
// Check true residual
///////////////////////////////////
    // t = D yd
#ifdef _CLOVER
    assign_mult_eoprec_wd(mkappa2d,ted,yded,ued,uod,ieod,fclinved,fclinvod);
#else
    assign_mult_eoprec_wd(mkappa2d,ted,yded,ued,uod,ieod);
#endif
    assign(red,yed);
    // r = r - t
    accum_sub(red,ted,ieod);
    rtmp0 = abs2(red,ieod);
    err = sqrt(rtmp0);
    printf("%%%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif

//======================================
// receive result from device to host
//======================================
#ifdef _DEBUG_
  cutStartTimer(yrecvtime);
#endif
  CUDA_SAFE_CALL(cudaMemcpy(yde,yded,sizeof(cuhqfield),cudaMemcpyDeviceToHost));
  yrecv_count += sizeof(cuhqfield);
#ifdef _DEBUG_
  cutStopTimer(yrecvtime);
#endif

//======================================
// deallocate device memory
//======================================
  CUDA_SAFE_CALL(cudaFree(ieod));
  CUDA_SAFE_CALL(cudaFree(ioed));

  CUDA_SAFE_CALL(cudaFree(alpd));
  CUDA_SAFE_CALL(cudaFree(nalpd));
  CUDA_SAFE_CALL(cudaFree(betd));
  CUDA_SAFE_CALL(cudaFree(omgd));
  CUDA_SAFE_CALL(cudaFree(nomgd));
  CUDA_SAFE_CALL(cudaFree(mkappa2d));

  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(yded));

  CUDA_SAFE_CALL(cudaFree(red));
  CUDA_SAFE_CALL(cudaFree(rted));
  CUDA_SAFE_CALL(cudaFree(ped));
  CUDA_SAFE_CALL(cudaFree(qed));
  CUDA_SAFE_CALL(cudaFree(ted));

#ifdef _DEBUG_
  fflush(stdout);
  cutStopTimer(solvtime);
#endif
}

extern "C" void 
__TEST_MULT(double *kappa, cuhqfield *ye, cuhqfield *dye, int *ieo)
{
  int ioe;
  int *ieod, *ioed;
  float *mkappa2d, mkappa2;
  cuhqfield *yed,*dyed;

  ioe = 1-*ieo;
  mkappa2 = -float((*kappa)*(*kappa));
  
  CUDA_SAFE_CALL(cudaMalloc((void**)&ieod,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&ioed,sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&mkappa2d,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**) &yed,sizeof(cuhqfield)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&dyed,sizeof(cuhqfield)));

  assign(*mkappa2d,mkappa2);

  CUDA_SAFE_CALL(cudaMemcpy(ieod, ieo,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(ioed,&ioe,sizeof(int),cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(yed,ye,sizeof(cuhqfield),cudaMemcpyHostToDevice));

  // dye = D ye
#ifdef _CLOVER
  assign_mult_eoprec_wd(mkappa2d,dyed,yed,ued,uod,ieod,fclinved,fclinvod);
#else
  assign_mult_eoprec_wd(mkappa2d,dyed,yed,ued,uod,ieod);
#endif

  CUDA_SAFE_CALL(cudaMemcpy(dye,dyed,sizeof(cuhqfield),cudaMemcpyDeviceToHost));

  CUDA_SAFE_CALL(cudaFree(dyed));
  CUDA_SAFE_CALL(cudaFree(yed));
  CUDA_SAFE_CALL(cudaFree(mkappa2d));
  CUDA_SAFE_CALL(cudaFree(ioed));
  CUDA_SAFE_CALL(cudaFree(ieod));
}

#endif
