static __global__ void 
accum_sub_blk_all( cu_wqf_eo_blk *re, const cu_wqf_eo_blk *qe){
//
// r = r - q
//
  unsigned int ithd = threadIdx.x;
  unsigned int iblk =  blockIdx.x;
#pragma unroll 6
  for (unsigned int i=0;i<CSH;i++) {
    (*re).blk[iblk].thd[i].f[ithd].x -= (*qe).blk[iblk].thd[i].f[ithd].x;
    (*re).blk[iblk].thd[i].f[ithd].y -= (*qe).blk[iblk].thd[i].f[ithd].y;
    (*re).blk[iblk].thd[i].f[ithd].z -= (*qe).blk[iblk].thd[i].f[ithd].z;
    (*re).blk[iblk].thd[i].f[ithd].w -= (*qe).blk[iblk].thd[i].f[ithd].w;
  }
}

static __global__ void 
accum_add_blk_all(cu_wqf_eo_blk *xe, const cu_wqf_eo_blk *ve){
//
// x = x + v
//
  unsigned int ithd = threadIdx.x;
  unsigned int iblk =  blockIdx.x;
#pragma unroll 6
  for (unsigned int i=0;i<CSH;i++) {
    (*xe).blk[iblk].thd[i].f[ithd].x += (*ve).blk[iblk].thd[i].f[ithd].x;
    (*xe).blk[iblk].thd[i].f[ithd].y += (*ve).blk[iblk].thd[i].f[ithd].y;
    (*xe).blk[iblk].thd[i].f[ithd].z += (*ve).blk[iblk].thd[i].f[ithd].z;
    (*xe).blk[iblk].thd[i].f[ithd].w += (*ve).blk[iblk].thd[i].f[ithd].w;
  }
}

static __global__ void 
accum_nrmult_add_blk_all( cu_wqf_eo_blk *yde, const cu_wqf_eo_blk *ye){
//
// yd = - yd * (-kappa2) + y
//
  float4 rri,yri;
  unsigned int ithd = threadIdx.x;
  unsigned int iblk =  blockIdx.x;
  float kappa2= - tex1Dfetch(tex_Float,0);

#pragma unroll 6
  for (unsigned int i=0;i<CSH;i++) {
    rri = (*yde).blk[iblk].thd[i].f[ithd];
    yri =  (*ye).blk[iblk].thd[i].f[ithd];
    yri.x += kappa2 * rri.x;
    yri.y += kappa2 * rri.y;
    yri.z += kappa2 * rri.z;
    yri.w += kappa2 * rri.w;
    (*yde).blk[iblk].thd[i].f[ithd] = yri;
  }

}


static void 
assign_inv_mult_eoprec_wd_eo_blk_all( const float *mkappa2,
                                    cu_wqf_eo_blk *yde,
                              const cu_wqf_eo_blk *ye,
                              const cu_gvf_eo_blk *ue,
                              const cu_gvf_eo_blk *uo,
                             const cu_clvf_eo_blk *fclinve = NULL,
                             const cu_clvf_eo_blk *fclinvo = NULL){
//
//
// yde = Dee \ ye  (in a block including extended sites)
//
//
  cu_wqf_eo_blk *Myo;

  CUDA_SAFE_CALL(cudaBindTexture(0,tex_Float,mkappa2,sizeof(float)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&(Myo),sizeof(cu_wqf_eo_blk)));

  // yde = ye
  assign(yde,ye);

  for (unsigned iter = 0; iter < _NMR; iter++) {

    // Myo = Moe yde
#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,yde,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri, uo,sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori, ue,sizeof(cu_gvf_eo_blk)));
#endif
    if (is_clover) {
      assign_mult_hop_kernel<<<_NBXYZ_Q_,_NTHREADS>>>(Myo,yde,uo,ue,fclinvo);
    } else {
      assign_mult_hop_kernel<<<_NBXYZ_Q_,_NTHREADS>>>(Myo,yde,uo,ue);
    }
#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

    //  ve = Meo Myo
#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,Myo,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri, ue,sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori, uo,sizeof(cu_gvf_eo_blk)));
#endif
    if (is_clover) {
      assign_mult_hop_kernel<<<_NBXYZ_Q_,_NTHREADS>>>(yde,Myo,ue,uo,fclinve);
    } else {
      assign_mult_hop_kernel<<<_NBXYZ_Q_,_NTHREADS>>>(yde,Myo,ue,uo);
    }
#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

    // yde = - yde * mkappa2 + ye
    accum_nrmult_add_blk_all<<<_NBXYZ_Q_,_NTHREADS>>>(yde,ye);
  }

  CUDA_SAFE_CALL(cudaFree(Myo));
  CUDA_SAFE_CALL(cudaUnbindTexture(tex_Float));

}

static void 
assign_inv_mult_eoprec_wd_ras(const float_gpu &mkappa2,
                                    cu_wqf_eo *yde, 
                              const cu_wqf_eo *ye, 
                                 const cu_gvf *u,
                                const cu_clvf *fclinv = NULL){
//
// Multiply even/odd-site preconditioned Dirac operator
//                                    RAS precondtioner: yde = Dee \ ye,
// where Dee = 1 - kappa^2 Meo Moe
//
// mkappa2 : negative hopping parameter squared (= -kappa^2)
//      ye : even/odd site fermion (input)
//     yde : even/odd site fermion (output)
//       u : gauge link
//  fclinv : inverse clover term
//

  cu_wqf_eo *re,*ve,*qe;
  int ieo = (*ye).ieo;
  int ioe = 1-ieo;
  (*yde).ieo  = ieo;
  (*yde).ipeo = _ipeo_;

  cutStartTimer(prec_timer);

  re = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  ve = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  qe = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  (*re).ieo = ieo;
  (*ve).ieo = ieo;
  (*qe).ieo = ieo;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    CUDA_SAFE_CALL(cudaMalloc((void **)&((*re).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&((*ve).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void **)&((*qe).g[igx]),sizeof(cu_wqf_eo_blk)));
    assign((*yde).g[igx],(*ye).g[igx]);
    assign( (*ve).g[igx],(*ye).g[igx]);
    assign( (*re).g[igx],(*ye).g[igx]);
//    CUDA_SAFE_CALL(cudaThreadSynchronize());
  }

  for (unsigned int iter=0;iter<_NRAS;iter++) {

    //  q = A v
    if (is_clover) {
      assign_mult_eoprec_wd_eo(mkappa2,qe,ve,u,fclinv);
    } else {
      assign_mult_eoprec_wd_eo(mkappa2,qe,ve,u);
    }


    //  r =  r - q
    cutStartTimer(binv_timer);
#pragma omp parallel for
    for (int igx=0;igx<_NGX;igx++) {
      accum_sub_blk_all<<<_NBXYZ_Q_,_NTHREADS>>>((*re).g[igx],(*qe).g[igx]);
//      CUDA_SAFE_CALL(cudaThreadSynchronize());
    }
    cutStopTimer(binv_timer);

    //  v = A' \ r
    // yd = yd + v

    copy_boundary_all(re); // copy boundary including extended sites.

    cutStartTimer(binv_timer);
#pragma omp parallel for
    for (int igx=0;igx<_NGX;igx++) {

//      CUDA_SAFE_CALL(cudaThreadSynchronize());

      if (is_clover) {
        assign_inv_mult_eoprec_wd_eo_blk_all(mkappa2.g[igx],
                                               (*ve).g[igx],
                                               (*re).g[igx],
                                        (*u).eo[ieo].g[igx],
                                        (*u).eo[ioe].g[igx],
                                   (*fclinv).eo[ieo].g[igx],
                                   (*fclinv).eo[ioe].g[igx]);
      } else {
        assign_inv_mult_eoprec_wd_eo_blk_all(mkappa2.g[igx],
                                               (*ve).g[igx],
                                               (*re).g[igx],
                                        (*u).eo[ieo].g[igx],
                                        (*u).eo[ioe].g[igx]);
      }

//      CUDA_SAFE_CALL(cudaThreadSynchronize());

      accum_add_blk_all<<<_NBXYZ_Q_,_NTHREADS>>>((*yde).g[igx],(*ve).g[igx]);

//      CUDA_SAFE_CALL(cudaThreadSynchronize());
    }
    cutStopTimer(binv_timer);

  }

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    CUDA_SAFE_CALL(cudaFree((*re).g[igx]));
    CUDA_SAFE_CALL(cudaFree((*ve).g[igx]));
    CUDA_SAFE_CALL(cudaFree((*qe).g[igx]));
  }
  _mm_free(re);
  _mm_free(ve);
  _mm_free(qe);

  cutStopTimer(prec_timer);

}

#define MULT_PREC ((MULT_MB_PRE + ((MULT_MB_PRE*_NMR+2*24.0)*_NBXYZ_Q_)/_PHYS_BLOCK_ )*_NRAS)

