static __device__ __inline__ void 
mult_hop_in_kernel(float ytr[],
                   float yti[],
                   const int &ithd,
                   const int &ibx,
                   const int &iby,
                   const int &ibz,
                   const int &iblk,
                   const int &igx,
                   const int &ieo,
                   const int &ipeo,
                   const cu_wqf_eo_blk *yo,
                   const cu_gvf_eo_blk *ue,
                   const cu_gvf_eo_blk *uo){


  { 
    int iz     = thd_z + ibz*_NTHZ;
    int iy     =         iby;
    int ix     =         ibx + (igx-1)*_NBX;
    int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
    int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
    int ithdtp = (((it+1+NT)%NT)+1)/2-ieoxyz;
    int ithdp  = ithd + (ithdtp % _NTHT - thd_t);
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[1] = __loady_tex(tex_yori,ithdp,1,iblk);
    yri[2] = __loady_tex(tex_yori,ithdp,2,iblk);
    yri[4] = __loady_tex(tex_yori,ithdp,4,iblk);
    yri[5] = __loady_tex(tex_yori,ithdp,5,iblk);
    uri[0] = __loadu_tex(tex_ueri,ithd ,0,3,iblk);
    uri[1] = __loadu_tex(tex_ueri,ithd ,1,3,iblk);
    uri[2] = __loadu_tex(tex_ueri,ithd ,2,3,iblk);
#else
    yri[1] = (*yo).blk[iblk].thd[1].f[ithdp];
    yri[2] = (*yo).blk[iblk].thd[2].f[ithdp];
    yri[4] = (*yo).blk[iblk].thd[4].f[ithdp];
    yri[5] = (*yo).blk[iblk].thd[5].f[ithdp];
    uri[0] = (*ue).blk[iblk].thd[3][0].f[ithd];
    uri[1] = (*ue).blk[iblk].thd[3][1].f[ithd];
    uri[2] = (*ue).blk[iblk].thd[3][2].f[ithd];
#endif
    cu_mult_forw_t(uri,yri,ytr,yti);
  }

  { 
    int iz     = thd_z + ibz*_NTHZ;
    int iy     =         iby;
    int ix     =         ibx + (igx-1)*_NBX;
    int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
    int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
    int ithdtd = (((it-1+NT)%NT)+1)/2-ieoxyz;
    int ithdd  = ithd + (ithdtd % _NTHT - thd_t);
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithdd,0,iblk);
    yri[1] = __loady_tex(tex_yori,ithdd,1,iblk);
    yri[3] = __loady_tex(tex_yori,ithdd,3,iblk);
    yri[4] = __loady_tex(tex_yori,ithdd,4,iblk);
    uri[0] = __loadu_tex(tex_uori,ithdd,0,3,iblk);
    uri[1] = __loadu_tex(tex_uori,ithdd,1,3,iblk);
    uri[2] = __loadu_tex(tex_uori,ithdd,2,3,iblk);
#else
    yri[0] = (*yo).blk[iblk].thd[0].f[ithdd];
    yri[1] = (*yo).blk[iblk].thd[1].f[ithdd];
    yri[3] = (*yo).blk[iblk].thd[3].f[ithdd];
    yri[4] = (*yo).blk[iblk].thd[4].f[ithdd];
    uri[0] = (*uo).blk[iblk].thd[3][0].f[ithdd];
    uri[1] = (*uo).blk[iblk].thd[3][1].f[ithdd];
    uri[2] = (*uo).blk[iblk].thd[3][2].f[ithdd];
#endif
    cu_mult_back_t(uri,yri,ytr,yti);
  }

  { 
    int iz     = thd_z + ibz*_NTHZ;
    int izp    = (iz+1+NZ)%NZ;
    int ithdzp = izp % _NTHZ;
    int iblkzp = izp / _NTHZ;
    int ithdp  = ithd + (ithdzp - thd_z)*_NTHT;
    int iblkp  = iblk + (iblkzp - blk_z(iblk));
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithdp,0,iblkp);
    yri[1] = __loady_tex(tex_yori,ithdp,1,iblkp);
    yri[2] = __loady_tex(tex_yori,ithdp,2,iblkp);
    yri[3] = __loady_tex(tex_yori,ithdp,3,iblkp);
    yri[4] = __loady_tex(tex_yori,ithdp,4,iblkp);
    yri[5] = __loady_tex(tex_yori,ithdp,5,iblkp);
    uri[0] = __loadu_tex(tex_ueri,ithd ,0,2,iblk);
    uri[1] = __loadu_tex(tex_ueri,ithd ,1,2,iblk);
    uri[2] = __loadu_tex(tex_ueri,ithd ,2,2,iblk);
#else
    yri[0] = (*yo).blk[iblkp].thd[0].f[ithdp];
    yri[1] = (*yo).blk[iblkp].thd[1].f[ithdp];
    yri[2] = (*yo).blk[iblkp].thd[2].f[ithdp];
    yri[3] = (*yo).blk[iblkp].thd[3].f[ithdp];
    yri[4] = (*yo).blk[iblkp].thd[4].f[ithdp];
    yri[5] = (*yo).blk[iblkp].thd[5].f[ithdp];
    uri[0] = (*ue).blk[iblk].thd[2][0].f[ithd];
    uri[1] = (*ue).blk[iblk].thd[2][1].f[ithd];
    uri[2] = (*ue).blk[iblk].thd[2][2].f[ithd];
#endif
    cu_mult_forw_z(uri,yri,ytr,yti);
  }

  { 
    int iz     = thd_z + ibz*_NTHZ;
    int izd    = (iz-1+NZ)%NZ;
    int ithdzd = izd % _NTHZ;
    int iblkzd = izd / _NTHZ;
    int ithdd  = ithd + (ithdzd - thd_z)*_NTHT;
    int iblkd  = iblk + (iblkzd - blk_z(iblk));
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithdd,0,iblkd);
    yri[1] = __loady_tex(tex_yori,ithdd,1,iblkd);
    yri[2] = __loady_tex(tex_yori,ithdd,2,iblkd);
    yri[3] = __loady_tex(tex_yori,ithdd,3,iblkd);
    yri[4] = __loady_tex(tex_yori,ithdd,4,iblkd);
    yri[5] = __loady_tex(tex_yori,ithdd,5,iblkd);
    uri[0] = __loadu_tex(tex_uori,ithdd,0,2,iblkd);
    uri[1] = __loadu_tex(tex_uori,ithdd,1,2,iblkd);
    uri[2] = __loadu_tex(tex_uori,ithdd,2,2,iblkd);
#else
    yri[0] = (*yo).blk[iblkd].thd[0].f[ithdd];
    yri[1] = (*yo).blk[iblkd].thd[1].f[ithdd];
    yri[2] = (*yo).blk[iblkd].thd[2].f[ithdd];
    yri[3] = (*yo).blk[iblkd].thd[3].f[ithdd];
    yri[4] = (*yo).blk[iblkd].thd[4].f[ithdd];
    yri[5] = (*yo).blk[iblkd].thd[5].f[ithdd];
    uri[0] = (*uo).blk[iblkd].thd[2][0].f[ithdd];
    uri[1] = (*uo).blk[iblkd].thd[2][1].f[ithdd];
    uri[2] = (*uo).blk[iblkd].thd[2][2].f[ithdd];
#endif
    cu_mult_back_z(uri,yri,ytr,yti);
  }

  { 
    int iblkp  = iblk + _NBZ;
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithd ,0,iblkp);
    yri[1] = __loady_tex(tex_yori,ithd ,1,iblkp);
    yri[2] = __loady_tex(tex_yori,ithd ,2,iblkp);
    yri[3] = __loady_tex(tex_yori,ithd ,3,iblkp);
    yri[4] = __loady_tex(tex_yori,ithd ,4,iblkp);
    yri[5] = __loady_tex(tex_yori,ithd ,5,iblkp);
    uri[0] = __loadu_tex(tex_ueri,ithd ,0,1,iblk);
    uri[1] = __loadu_tex(tex_ueri,ithd ,1,1,iblk);
    uri[2] = __loadu_tex(tex_ueri,ithd ,2,1,iblk);
#else
    yri[0] = (*yo).blk[iblkp].thd[0].f[ithd];
    yri[1] = (*yo).blk[iblkp].thd[1].f[ithd];
    yri[2] = (*yo).blk[iblkp].thd[2].f[ithd];
    yri[3] = (*yo).blk[iblkp].thd[3].f[ithd];
    yri[4] = (*yo).blk[iblkp].thd[4].f[ithd];
    yri[5] = (*yo).blk[iblkp].thd[5].f[ithd];
    uri[0] = (*ue).blk[iblk].thd[1][0].f[ithd];
    uri[1] = (*ue).blk[iblk].thd[1][1].f[ithd];
    uri[2] = (*ue).blk[iblk].thd[1][2].f[ithd];
#endif
    cu_mult_forw_y(uri,yri,ytr,yti);
  }

  { 
    int iblkd  = iblk - _NBZ;
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithd ,0,iblkd);
    yri[1] = __loady_tex(tex_yori,ithd ,1,iblkd);
    yri[2] = __loady_tex(tex_yori,ithd ,2,iblkd);
    yri[3] = __loady_tex(tex_yori,ithd ,3,iblkd);
    yri[4] = __loady_tex(tex_yori,ithd ,4,iblkd);
    yri[5] = __loady_tex(tex_yori,ithd ,5,iblkd);
    uri[0] = __loadu_tex(tex_uori,ithd ,0,1,iblkd);
    uri[1] = __loadu_tex(tex_uori,ithd ,1,1,iblkd);
    uri[2] = __loadu_tex(tex_uori,ithd ,2,1,iblkd);
#else
    yri[0] = (*yo).blk[iblkd].thd[0].f[ithd];
    yri[1] = (*yo).blk[iblkd].thd[1].f[ithd];
    yri[2] = (*yo).blk[iblkd].thd[2].f[ithd];
    yri[3] = (*yo).blk[iblkd].thd[3].f[ithd];
    yri[4] = (*yo).blk[iblkd].thd[4].f[ithd];
    yri[5] = (*yo).blk[iblkd].thd[5].f[ithd];
    uri[0] = (*uo).blk[iblkd].thd[1][0].f[ithd];
    uri[1] = (*uo).blk[iblkd].thd[1][1].f[ithd];
    uri[2] = (*uo).blk[iblkd].thd[1][2].f[ithd];
#endif
    cu_mult_back_y(uri,yri,ytr,yti);
  }

  { 
    int iblkp  = iblk + _NBZ*(_NBY+2*_NDEPTH);
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithd ,0,iblkp);
    yri[1] = __loady_tex(tex_yori,ithd ,1,iblkp);
    yri[2] = __loady_tex(tex_yori,ithd ,2,iblkp);
    yri[3] = __loady_tex(tex_yori,ithd ,3,iblkp);
    yri[4] = __loady_tex(tex_yori,ithd ,4,iblkp);
    yri[5] = __loady_tex(tex_yori,ithd ,5,iblkp);
    uri[0] = __loadu_tex(tex_ueri,ithd ,0,0,iblk);
    uri[1] = __loadu_tex(tex_ueri,ithd ,1,0,iblk);
    uri[2] = __loadu_tex(tex_ueri,ithd ,2,0,iblk);
#else
    yri[0] = (*yo).blk[iblkp].thd[0].f[ithd];
    yri[1] = (*yo).blk[iblkp].thd[1].f[ithd];
    yri[2] = (*yo).blk[iblkp].thd[2].f[ithd];
    yri[3] = (*yo).blk[iblkp].thd[3].f[ithd];
    yri[4] = (*yo).blk[iblkp].thd[4].f[ithd];
    yri[5] = (*yo).blk[iblkp].thd[5].f[ithd];
    uri[0] = (*ue).blk[iblk].thd[0][0].f[ithd];
    uri[1] = (*ue).blk[iblk].thd[0][1].f[ithd];
    uri[2] = (*ue).blk[iblk].thd[0][2].f[ithd];
#endif
    cu_mult_forw_x(uri,yri,ytr,yti);
  }

  { 
    int iblkd  = iblk - _NBZ*(_NBY+2*_NDEPTH);
    __align__(16) float4 yri[CSH];
    __align__(16) float4 uri[COL];
#ifdef _USE_TEXFETCH_
    yri[0] = __loady_tex(tex_yori,ithd ,0,iblkd);
    yri[1] = __loady_tex(tex_yori,ithd ,1,iblkd);
    yri[2] = __loady_tex(tex_yori,ithd ,2,iblkd);
    yri[3] = __loady_tex(tex_yori,ithd ,3,iblkd);
    yri[4] = __loady_tex(tex_yori,ithd ,4,iblkd);
    yri[5] = __loady_tex(tex_yori,ithd ,5,iblkd);
    uri[0] = __loadu_tex(tex_uori,ithd ,0,0,iblkd);
    uri[1] = __loadu_tex(tex_uori,ithd ,1,0,iblkd);
    uri[2] = __loadu_tex(tex_uori,ithd ,2,0,iblkd);
#else
    yri[0] = (*yo).blk[iblkd].thd[0].f[ithd];
    yri[1] = (*yo).blk[iblkd].thd[1].f[ithd];
    yri[2] = (*yo).blk[iblkd].thd[2].f[ithd];
    yri[3] = (*yo).blk[iblkd].thd[3].f[ithd];
    yri[4] = (*yo).blk[iblkd].thd[4].f[ithd];
    yri[5] = (*yo).blk[iblkd].thd[5].f[ithd];
    uri[0] = (*uo).blk[iblkd].thd[0][0].f[ithd];
    uri[1] = (*uo).blk[iblkd].thd[0][1].f[ithd];
    uri[2] = (*uo).blk[iblkd].thd[0][2].f[ithd];
#endif
    cu_mult_back_x(uri,yri,ytr,yti);
  }
}

//
// For wilson fermion
//
static __global__ void 
assign_mult_hop_in_kernel(      cu_wqf_eo_blk *yde,
                          const int4          *blk_indx, 
                          const cu_wqf_eo_blk *yo, 
                          const cu_gvf_eo_blk *ue,
                          const cu_gvf_eo_blk *uo){
//
//  Multiply odd->even/even->odd hopping matrix : yde = Meo yo
// 
//      yde : even/odd site fermion 
//       yo : odd/even site fermion 
//       ue : even/odd site link
//       uo : odd/even site link
// 
  __align__(16) float ytr[CLSP],yti[CLSP];
  int ithd = threadIdx.x;
  int ibx  = blk_indx[blockIdx.x].x;
  int iby  = blk_indx[blockIdx.x].y;
  int ibz  = blk_indx[blockIdx.x].z;
  int iblk = blk_indx[blockIdx.x].w;
  int igx  =   (*yo).igx;
  int ieo  = 1-(*yo).ieo;
  int ipeo =   (*yo).ipeo;

  (*yde).igx  = igx;
  (*yde).ieo  = ieo;
  (*yde).ipeo = ipeo;

  ytr[0]=0.0f;
  ytr[1]=0.0f;
  ytr[2]=0.0f;
  ytr[3]=0.0f;
  ytr[4]=0.0f;
  ytr[5]=0.0f;
  yti[0]=0.0f;
  yti[1]=0.0f;
  yti[2]=0.0f;
  yti[3]=0.0f;
  yti[4]=0.0f;
  yti[5]=0.0f;

  mult_hop_in_kernel(ytr,yti,ithd,ibx,iby,ibz,iblk,igx,ieo,ipeo,yo,ue,uo);

#ifdef _SF
  {
//
//  Set SF boundary zero (Fortran: itt=1 (C: itt=0))
//
    int iz     = thd_z + ibz*_NTHZ;
    int iy     =         iby;
    int ix     =         ibx + (igx-1)*_NBX;
    int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
    int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
    float flag = (float)(1-(_NTT-it)/_NTT);
#pragma unroll 12
    for (unsigned int i = 0; i < CLSP; i++) {
      ytr[i] = ytr[i] * flag;
      yti[i] = yti[i] * flag;
    } 
  }
#endif

  (*yde).blk[iblk].thd[0].f[ithd] = *(float4 *)(ytr+0);
  (*yde).blk[iblk].thd[1].f[ithd] = *(float4 *)(ytr+4);
  (*yde).blk[iblk].thd[2].f[ithd] = *(float4 *)(ytr+8);
  (*yde).blk[iblk].thd[3].f[ithd] = *(float4 *)(yti+0);
  (*yde).blk[iblk].thd[4].f[ithd] = *(float4 *)(yti+4);
  (*yde).blk[iblk].thd[5].f[ithd] = *(float4 *)(yti+8);

}

//
// For clvoer fermion
//
static __global__ void 
assign_mult_hop_in_kernel(      cu_wqf_eo_blk *yde,
                          const int4          *blk_indx, 
                          const cu_wqf_eo_blk *yo, 
                          const cu_gvf_eo_blk *ue,
                          const cu_gvf_eo_blk *uo,
                         const cu_clvf_eo_blk *fclinve) {
//
//
// Multiply odd->even/even->odd hopping matrix : yde = Fee \ Meo yo
//
//       yo : odd/even site fermion
//      yde : even/odd site fermion
//       ue : even/odd site link
//       uo : odd/even site link
//  fclinve : even/odd site inverse clover term
// 
  __align__(16) float ytr[CLSP],yti[CLSP];
  int ithd = threadIdx.x;
  int ibx  = blk_indx[blockIdx.x].x;
  int iby  = blk_indx[blockIdx.x].y;
  int ibz  = blk_indx[blockIdx.x].z;
  int iblk = blk_indx[blockIdx.x].w;
  int igx  =   (*yo).igx;
  int ieo  = 1-(*yo).ieo;
  int ipeo =   (*yo).ipeo;

  (*yde).igx  = igx;
  (*yde).ieo  = ieo;
  (*yde).ipeo = ipeo;

  ytr[0]=0.0f;
  ytr[1]=0.0f;
  ytr[2]=0.0f;
  ytr[3]=0.0f;
  ytr[4]=0.0f;
  ytr[5]=0.0f;
  yti[0]=0.0f;
  yti[1]=0.0f;
  yti[2]=0.0f;
  yti[3]=0.0f;
  yti[4]=0.0f;
  yti[5]=0.0f;

  mult_hop_in_kernel(ytr,yti,ithd,ibx,iby,ibz,iblk,igx,ieo,ipeo,yo,ue,uo);

  cu_mult_clover_term(&((*fclinve).blk[iblk]),&ithd,ytr,yti);

#ifdef _SF
  {
//
//  Set SF boundary zero (Fortran: itt=1 (C: itt=0))
//
    int iz     = thd_z + ibz*_NTHZ;
    int iy     =         iby;
    int ix     =         ibx + (igx-1)*_NBX;
    int ieoxyz = (ipeo+ieo+(iz+1)+(iy+1)+(ix+1)+256)%2;
    int it     = (thd_t+1-ieoxyz)*2+ieoxyz-1;
    float flag = (float)(1-(_NTT-it)/_NTT);
#pragma unroll 12
    for (unsigned int i = 0; i < CLSP; i++) {
      ytr[i] = ytr[i] * flag;
      yti[i] = yti[i] * flag;
    } 
  }
#endif

  (*yde).blk[iblk].thd[0].f[ithd] = *(float4 *)(ytr+0);
  (*yde).blk[iblk].thd[1].f[ithd] = *(float4 *)(ytr+4);
  (*yde).blk[iblk].thd[2].f[ithd] = *(float4 *)(ytr+8);
  (*yde).blk[iblk].thd[3].f[ithd] = *(float4 *)(yti+0);
  (*yde).blk[iblk].thd[4].f[ithd] = *(float4 *)(yti+4);
  (*yde).blk[iblk].thd[5].f[ithd] = *(float4 *)(yti+8);

}
