static cuComplex
assign_neg(cuComplex &a){
//
// c = -a
//
  cuComplex c;
  c.x = -a.x;
  c.y = -a.y;
  return c;
}

static cuComplex
assign_mul(cuComplex &a, cuComplex &b){
//
// c = a * b
//
  cuComplex c;
  c.x = a.x * b.x - a.y * b.y;
  c.y = a.x * b.y + a.y * b.x;
  return c;
}

static cuComplex
assign_div(cuComplex &a, cuComplex &b){
//
// c = a/b
//
  cuComplex c;
  float den,bri;
  if (abs(b.x) > abs(b.y)) {
    bri = b.y / b.x;
    c.x = a.x + a.y * bri;
    c.y = a.y - a.x * bri;
    den = b.x + b.y * bri;
  } else {
    bri = b.x / b.y;
    c.x = a.x * bri + a.y;
    c.y = a.y * bri - a.x;
    den = b.x * bri + b.y;
  }
  c.x = c.x / den;
  c.y = c.y / den;
  return c;
} 

static cuComplex
assign_div(cuComplex &a, float &b){
//
// c = a/b
//
  cuComplex c;
  c.x = a.x / b;
  c.y = a.y / b;
  return c;
}

static void cuPutRealScalar(const float &a, float &c){
  CUDA_SAFE_CALL(cudaMemcpy(&c, &a, sizeof(float), cudaMemcpyHostToDevice));
}

static void cuPutComplexScalar(const cuComplex &a, cuComplex &c){
  CUDA_SAFE_CALL(cudaMemcpy(&c, &a, sizeof(cuComplex), cudaMemcpyHostToDevice));
}

static void assign(float_gpu &c, const float &a){
//
// c = a
//
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    cuPutRealScalar(a, *(c.g[igx]));
  }
}

static void assign(cuComplex_gpu &c, const cuComplex &a){
//
// c = a
//
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
    cuPutComplexScalar(a, *(c.g[igx]));
  }
}

