#include <config.h>
module qprop_class
  use lattice_class
  use error_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use hmc_std_quark_wilson_class
  use quark_dw_ovlp_class
  use wavefunction_class
  implicit none
  private

  public :: quark_prop
  public :: new
  public :: delete
  public :: get

  type :: quark_prop
    complex(DP) :: qprop(COL,COL,SPIN,SPIN,NX*NY*NZ,NT)
!    type(quark_truncd_dw_overlap) :: parameter
    class(quark_wilson), pointer :: parameter => NULL()
    type(wavefunc), pointer :: wf => NULL()
    integer :: iout,iiter
    real(DP) :: tol
  end type

  interface new
    module procedure new_qprop
  end interface

  interface delete
    module procedure delete_qprop
  end interface

  interface get
    module procedure get_qprop
  end interface

contains

subroutine new_qprop(this,iout)
  use file_tools_class
  implicit none
  type(quark_prop), intent(inout) :: this
  integer, optional, intent(in) :: iout

  if (present(iout)) then
    this%iout = iout
  else
    this%iout = search_free_file_unit()
  endif
  this%tol   = 1.0e-14_DP
  this%iiter = 10000

  return
end subroutine

subroutine delete_qprop(this)
  implicit none
  type(quark_prop), intent(inout) :: this
  return
end subroutine

subroutine get_qprop(this,u)
  implicit none
  type(quark_prop),      intent(inout) :: this
  type(vfield_gluon_wg), intent(in) :: u
  type(field_quark_wg), allocatable ::  y
  type(field_quark_wg), allocatable :: dy
  integer :: ix,iy,iz,it,is,ic,js,jc,ixyz
  integer :: ieo,ieoxyz,ith
  integer :: iout,iiter
  real(DP) :: tol

  allocate(y,dy)
  call new(y)
  call new(dy)

  if (.not.associated(this%wf)) then
    call error_stop("please assing wavefunction pointer for qprop.")
  endif

  do js=1,SPIN
  do jc=1,COL

    !
    ! set local source
    !
    if (0==nodeid) then
      select case(this%wf%type)
      case(TYPE_WF_LOCAL)
        write(*,'("%M Set source TYPE= LOCAL")')
      case(TYPE_WF_JACOBI)
        write(*,'("%M Set source TYPE= JACOBI, Nsmr=",I3," Ksrm=",E24.15)')this%wf%Nsmr,this%wf%Ksmr
      end select
      write(*,'("%M (col,spin)=(",I3,",",I3,") at ")',advance='no') jc,js
      write(*,'("(it,iz,iy,ix)=(",I3,",",I3,",",I3,",",I3,")")') &
 &                              this%wf%itt0,this%wf%itz0,this%wf%ity0,this%wf%itx0
    endif

    call set_source(this%wf,jc,js,y,u)

    iout  = this%iout
    tol   = this%tol
    iiter = this%iiter
    if (0==nodeid) write(iout,'("# iout=",I3," tol=",E24.15," maxiter=",I6)')iout,tol,iiter

    select type(parameter => this%parameter)
    type is (quark_truncd_dw_overlap)
      call assign_inv_mult_ov(parameter,iout,tol,iiter,dy,y,u)
#ifdef _CLOVER
    type is (quark_clover)
#else
    type is (quark_wilson) 
#endif
      call assing_inv_mult_wd(parameter,iout,tol,iiter,dy,y,u)
    end select

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ixyz,it,ieo,ith,is,ic)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ixyz = iz + (iy-1)*NZ + (ix-1)*NZ*NY
    do it=1,NT
      ieo = mod(ipeo+it+iz+iy+ix,2)
      ith = it/2
      do is=1,SPIN
      do ic=1,COL
        this%qprop(ic,jc,is,js,ixyz,it) = dy%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is)
      enddo
      enddo
    enddo
    enddo
    enddo
    enddo

  enddo
  enddo

  deallocate(y,dy)

  return
end subroutine

subroutine assing_inv_mult_wd(param,iout,tol,iiter,dy,y,u)
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: param
#else
  type(quark_wilson), intent(inout) :: param
#endif
  integer,  intent(in)    :: iout
  real(DP), intent(in)    :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg),  intent(in)    :: y
  type(field_quark_wg),  intent(inout) :: dy
  type(vfield_gluon_wg), intent(in)    :: u
  type(field_quark_wg),  allocatable :: w
#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable :: wl34,wl98
#endif
  real(DP) :: dummy,rtmp
  integer :: ieo,ioe
  ieo = 0
  ioe = 1-ieo

#ifdef _CLOVER
  allocate(wl34,wl98)
  call make_two_inks(u,wl34,wl98)
  call make_clover_leaf(u,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(param,0,dummy)
  call delete_clover_leaf()
#endif

  allocate(w)
  call new(w)
#ifdef _CLOVER
  call assign_mult_clover_term(w%eo(ieo),param,y%eo(ieo))
  call assign_mult_clover_term(w%eo(ioe),param,y%eo(ioe))
#else
  call assign(w,y) ! w = y
#endif

  !
  !  we = we - Meo yo
  !
  call assign_mult_hop(param,dy%eo(ieo),w%eo(ioe),u)
  rtmp = - get_kappa(param)
  call accum_add_mult(w%eo(ieo),dy%eo(ieo),rtmp)

  call assign_inv_mult_eoprec_wd(param,iout,tol,iiter,dy%eo(ieo),w%eo(ieo),u)

  !
  !  dyo = wo - Moe dye
  !
  call assign(dy%eo(ioe),w%eo(ioe))
  call assign_mult_hop(param,w%eo(ioe),dy%eo(ieo),u)
  call accum_add_mult(dy%eo(ioe),w%eo(ioe),rtmp)

  deallocate(w)
#ifdef _CLOVER
  call delete_inverse_clover_term(param,0,dummy)
#endif

  return
end subroutine

end module
