#!/usr/bin/ruby
#
# Generates SSOR Kernel for Hermitian Squared Domainwall operator (D'D)
#
# $Id: GenMultSSOR.rb,v 1.3 2010/09/10 06:25:10 ishikawa Exp $
#

puts "xxxA"

require 'optparse'
require 'FieldClasses'

state = SiteRestriction.new("","blocknormal")

flag = 0

OptionParser.new {|opt|
  opt.on('--dagger') {
    flag = 1
  }
  opt.on('--normal') {
    flag = 0
  }
  opt.parse!(ARGV)
}

if (flag == 1) 
  puts "!======================="
  puts "! Dagger mode"
  puts "!======================="
else
  puts "!======================="
  puts "! Normal mode"
  puts "!======================="
end

amu = Array.new(4)
for i in 1..4
   amu[i-1] = i
end

puts "!================================================="

yt = Spinor.new("yt(iw)%y")
y  = Spinor.new("y")
w  = Spinor.new("w")

b  = FieldDWSpinor.new("q")
x  = FieldDWSpinor.new("dq")
vu = FieldGauge.new("vu","vector")

##########################################
#  yt = -kappa K \ Hop dq
##########################################

##################
#  yt = Hop4D dq
##################
puts "  do iw=1,NS"
puts "    yt(iw)%y(:,:) = Z0"
puts "  enddo"
for mu in amu.reverse
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"

  if (flag == 1)
    gmu = -mu
  else
    gmu =  mu
  end

  puts "  do iw=1,NS"
  ##########################################
  #  yt = yt + (1-gamma_mu)U_mu(n) x(n+mu)
  ##########################################
  puts  y.assign_proj_up(-gmu,x.field_shift(+mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field,y,+1)
  puts yt.accum_proj_low(-gmu,w)
  puts "  enddo"

  ##############################################
  #  yt = yt + (1+gamma_mu)U_mu(n-mu)^+ x(n-mu)
  ##############################################
  puts "  do iw=1,NS"
  puts  y.assign_proj_up(+gmu,x.field_shift(-mu))
  puts  w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,-1)
  puts yt.accum_proj_low(+gmu,w)
  puts "  enddo"

end

#########################################
# multply 5th-dim mass matrix inverse
#  yt = K \ yt
#########################################
print "
  do iw=1,NS
    yy(:,1,iw) = (yt(iw)%y(:,1) - yt(iw)%y(:,3))*0.5_DP
    yy(:,2,iw) = (yt(iw)%y(:,2) - yt(iw)%y(:,4))*0.5_DP
    ww(:,1,iw) = (yt(iw)%y(:,1) + yt(iw)%y(:,3))*0.5_DP
    ww(:,2,iw) = (yt(iw)%y(:,2) + yt(iw)%y(:,4))*0.5_DP
  enddo
  do iw=1,NS
    xy(:,:,iw) = z0
    xw(:,:,iw) = z0
"
if (flag == 1)
# normal mode  XR \ PL + XL \ PR
print "
    do jw=1,NS
      xy(:,:,iw) = xy(:,:,iw) + m_invXR(iw,jw)*yy(:,:jw) 
      xw(:,:,iw) = xw(:,:,iw) + m_invXL(iw,jw)*ww(:,:jw) 
    enddo
"
else
# normal mode  XL \ PL + XR \ PR
print "
    do jw=1,NS
      xy(:,:,iw) = xy(:,:,iw) + m_invXL(iw,jw)*yy(:,:jw) 
      xw(:,:,iw) = xw(:,:,iw) + m_invXR(iw,jw)*ww(:,:jw) 
    enddo
"
end
print "
    yt(iw)%y(:,1) =  xy(:,1,iw) + xw(:,1,iw)
    yt(iw)%y(:,2) =  xy(:,2,iw) + xw(:,2,iw)
    yt(iw)%y(:,3) = -xy(:,1,iw) + xw(:,1,iw)
    yt(iw)%y(:,4) = -xy(:,2,iw) + xw(:,2,iw)
  enddo
"

#################################################
# overrelaxation
# x = b + yt*kappa*omega
#################################################
puts ""
puts "  do iw=1,NS"
puts "    " + x.field.get_name + "(:,:) = " + b.field.get_name + "(:,:) + yt(iw)%y(:,:)*kappa*omega"
puts "  enddo"
puts ""
