#include <config.h>
module hadron_class
  use lattice_class
  use timer_class
  use comlib
  use quark_prop_class
  use wilson_quark_parameter_class
  implicit none
  private

  public :: new,delete
  public :: hadrons_obj
  public :: meson_prop_obj
  public :: meson_mom_prop_obj
  public :: baryon_prop_obj
  public :: get_prop, save
  public :: NUM_MESON_OP
  public :: NUM_MESON_OP_MOM,NUM_MOM
  public :: NUM_BARYON_OP
  public :: NUM_DELTA_BARYON
  public :: NUM_OCTET_BARYON

  integer, save :: io_unit=100

  integer, parameter :: NUM_DELTA_BARYON=4
  integer, parameter :: NUM_OCTET_BARYON=2
  integer, parameter :: NUM_BARYON_OP=NUM_DELTA_BARYON*2+4*NUM_OCTET_BARYON

  integer, parameter :: NUM_MESON_OP=16+2+9+3
  integer, parameter :: NUM_MESON_OP_LOCAL=18
  integer, parameter :: NUM_MESON_OP_MOM=4
  integer, parameter :: NUM_MOM=19
  integer, parameter :: NSPACE=NX*NY*NZ
  complex(8) :: zphi(NSPACE,NUM_MOM)   ! mementum phase table
  integer :: ig5a(SPIN,0:15),is5a(SPIN,0:15)     ! current operator gamma matrix index
  integer ::  i5a(0:15),ia5(0:15)                ! current operator gamma matrix phase
  complex(8) :: zi(0:3)                          ! complex number
  data zi /(1.d0,0.d0),(0.d0,1.d0),(-1.d0,0.d0),(0.d0,-1.d0)/
  data ig5a / 2,3,0,1,   &
 &            0,1,2,3,   &

 &            1,0,3,2,   &
 &            1,0,3,2,   &
 &            0,1,2,3,   &
 &            2,3,0,1,   &

 &            3,2,1,0,   &
 &            3,2,1,0,   &
 &            2,3,0,1,   &
 &            0,1,2,3,   &

 &            2,3,0,1,   &
 &            3,2,1,0,   &
 &            1,0,3,2,   &
 &            3,2,1,0,   &
 &            1,0,3,2,   &
 &            0,1,2,3 /
  data is5a / 1, 1, 1, 1,   &
 &            1, 1, 1, 1,   &
 &            1, 1,-1,-1,   &
 &            1,-1,-1, 1,   &
 &            1,-1,-1, 1,   &
 &            1, 1,-1,-1,   &
 &            1, 1,-1,-1,   &
 &            1,-1,-1, 1,   &
 &            1,-1,-1, 1,   &
 &            1, 1,-1,-1,   &
 &            1,-1, 1,-1,   &
 &            1,-1, 1,-1,   &
 &            1, 1, 1, 1,   &
 &            1, 1, 1, 1,   &
 &            1,-1, 1,-1,   &
 &            1,-1, 1,-1 /
  data i5a /   2,0,  &
 &             3,2,3,0,  &
 &             0,3,0,1,  &
 &             0,1,0,0,3,0 /
  data ia5 /   2,0,  &
 &             1,0,1,2,  &
 &             2,1,2,3,  &
 &             0,1,0,0,3,0 /
  integer :: color(0:35,2),icG(0:3,0:2),scG(0:3,0:2)
  integer :: ic5(0:3),sc5(0:3)
  data color /4,5,3,7,8,6,1,2,0,8,6,7,2,0,1,5,3,4,  &
 &            7,8,6,1,2,0,4,5,3,5,3,4,8,6,7,2,0,1,  &
 &            8,6,7,2,0,1,5,3,4,4,5,3,7,8,6,1,2,0,  &
 &            5,3,4,8,6,7,2,0,1,7,8,6,1,2,0,4,5,3/
  data icG /0,1,2,3,  &
 &          0,1,2,3,  &
 &          1,0,3,2/
  data scG /-1, 0, 1, 0,  &
 &           0, 1, 0,-1,  &
 &           1, 1,-1,-1/
  data ic5 /1,0,3,2/
  data sc5 /1,-1,1,-1/

  type hadrons_obj
    integer :: idummy(4)
  end type

  type baryon_prop_obj
    complex(8) :: prop(NTT,0:NUM_BARYON_OP-1)
    type(wilson_quark_parameter_obj) :: quark_mass(3)
    character(len=_CHARLEN_) :: fname="hpropB"
    character(len=_CHARLEN_) :: path
    type(timer_obj) :: run_time
    integer :: smear_sous(4)
    integer :: smear_sink(4)
    integer :: iout
    integer :: itraj
    integer :: idummy(2)
  end type

  type meson_prop_obj
    complex(8) :: prop(NTT,0:NUM_MESON_OP-1)
    type(wilson_quark_parameter_obj) :: quark_mass(2)
    character(len=_CHARLEN_) :: fname="hpropM"
    character(len=_CHARLEN_) :: path
    type(timer_obj) :: run_time
    integer :: smear_sous(4)
    integer :: smear_sink(4)
    integer :: iout
    integer :: itraj
    integer :: idummy(2)
  end type

  type meson_mom_prop_obj
    complex(8) :: prop(NTT,1:NUM_MESON_OP_MOM,1:NUM_MOM)
    type(wilson_quark_parameter_obj) :: quark_mass(2)
    character(len=_CHARLEN_) :: fname="hpropMP"
    character(len=_CHARLEN_) :: path
    type(timer_obj) :: run_time
    integer :: smear_sous(4)
    integer :: smear_sink(4)
    integer :: iout
    integer :: itraj
    integer :: idummy(2)
  end type

  interface new
    module procedure  new_hadron
    module procedure  new_meson
    module procedure  new_meson_mom
    module procedure  new_baryon
  end interface

  interface delete
    module procedure  delete_meson
    module procedure  delete_meson_mom
    module procedure  delete_baryon
  end interface

  interface get_prop
    module procedure  get_meson_prop
    module procedure  get_meson_mom_prop
    module procedure  get_baryon1_prop
    module procedure  get_baryon2C_prop
  end interface

  interface save
    module procedure  save_meson_prop
    module procedure  save_meson_mom_prop
    module procedure  save_baryon_prop
  end interface

contains

subroutine new_hadron(this)
  implicit none
  type(hadrons_obj), intent(inout) :: this

  call set_momentum_table
  this%idummy(:)=1

  return
end subroutine

subroutine new_meson(this)
  implicit none
  type(meson_prop_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)
  character(len=_CHARLEN_) :: fname
  logical :: flag

  call new(this%run_time)
  this%prop=z0

  if (nodeid==0) then
    do
      inquire(unit=io_unit,opened=flag)
      if (flag) then
        io_unit=io_unit+1
      else
        this%iout=io_unit
        exit
      endif
    enddo
  endif

  call comlib_bcast(this%iout,0)

  write(fname,'(I6.6)')this%itraj
  fname=TRIM(this%path)//'/'//TRIM(this%fname)//'.'//TRIM(fname)
  if (nodeid==0) open(unit=this%iout,file=TRIM(fname),form='formatted',status='new')

  return
end subroutine

subroutine new_meson_mom(this)
  implicit none
  type(meson_mom_prop_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)
  character(len=_CHARLEN_) :: fname
  logical :: flag

  call new(this%run_time)
  this%prop=z0

  if (nodeid==0) then
    do
      inquire(unit=io_unit,opened=flag)
      if (flag) then
        io_unit=io_unit+1
      else
        this%iout=io_unit
        exit
      endif
    enddo
  endif

  call comlib_bcast(this%iout,0)

  write(fname,'(I6.6)')this%itraj
  fname=TRIM(this%path)//'/'//TRIM(this%fname)//'.'//TRIM(fname)
  if (nodeid==0) open(unit=this%iout,file=TRIM(fname),form='formatted',status='new')

  return
end subroutine

subroutine new_baryon(this)
  implicit none
  type(baryon_prop_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)
  character(len=_CHARLEN_) :: fname
  logical :: flag

  call new(this%run_time)
  this%prop=z0

  if (nodeid==0) then
    do
      inquire(unit=io_unit,opened=flag)
      if (flag) then
        io_unit=io_unit+1
      else
        this%iout=io_unit
        exit
      endif
    enddo
  endif
  call comlib_bcast(this%iout,0)

  write(fname,'(I6.6)')this%itraj
  fname=TRIM(this%path)//'/'//TRIM(this%fname)//'.'//TRIM(fname)
  if (nodeid==0) open(unit=this%iout,file=TRIM(fname),form='formatted',status='new')

  return
end subroutine

subroutine delete_meson(this)
  implicit none
  type(meson_prop_obj), intent(inout) :: this

  if (nodeid==0) close(unit=this%iout)

  return
end subroutine

subroutine delete_meson_mom(this)
  implicit none
  type(meson_mom_prop_obj), intent(inout) :: this

  if (nodeid==0) close(unit=this%iout)

  return
end subroutine

subroutine delete_baryon(this)
  implicit none
  type(baryon_prop_obj), intent(inout) :: this

  if (nodeid==0) close(unit=this%iout)

  return
end subroutine

subroutine save_meson_prop(this,iflv)
  implicit none
  type(meson_prop_obj), intent(inout) :: this
  integer, intent(in) :: iflv
  integer :: itt,iop

  if (nodeid==0) then
    write(this%iout,'(I3,2X,2F15.8,2X,2I3)')iflv, &
 &                      this%quark_mass(1)%kappa, &
 &                      this%quark_mass(2)%kappa, &
 &                      this%smear_sous(1),       &
 &                      this%smear_sous(2)
    do iop=0,NUM_MESON_OP-1
      write(this%iout,'("op. ",I3," begin")')iop
      do itt=1,NTT
        write(this%iout,'(I3,2X,E15.8,2X,E15.8)')itt-1,this%prop(itt,iop)
      enddo
      write(this%iout,'("op. ",I3," end")')iop
    enddo
  endif

  return
end subroutine

subroutine save_meson_mom_prop(this,iflv)
  implicit none
  type(meson_mom_prop_obj), intent(inout) :: this
  integer, intent(in) :: iflv
  integer :: itt,iop,imom

  if (nodeid==0) then
    write(this%iout,'(I3,2X,2F15.8,2X,2I3)')iflv, &
 &                      this%quark_mass(1)%kappa, &
 &                      this%quark_mass(2)%kappa, &
 &                      this%smear_sous(1),       &
 &                      this%smear_sous(2)
    do iop=1,NUM_MESON_OP_MOM
      write(this%iout,'("op. ",I3," begin")')iop
      do imom=1,NUM_MOM
        write(this%iout,'("%momentum: ",I3," start")')imom
        do itt=1,NTT
          write(this%iout,'(I3,2X,E15.8,2X,E15.8)')itt-1,this%prop(itt,iop,imom)
        enddo
        write(this%iout,'("%momentum: ",I3," finish")')imom
      enddo
      write(this%iout,'("op. ",I3," end")')iop
    enddo
  endif

  return
end subroutine

subroutine save_baryon_prop(this,iflv)
  implicit none
  type(baryon_prop_obj), intent(inout) :: this
  integer, intent(in) :: iflv
  integer :: itt,iop

  if (nodeid==0) then
    write(this%iout,'(I3,2X,3F15.8,3X,3I3)')iflv, &
 &                      this%quark_mass(1)%kappa, &
 &                      this%quark_mass(2)%kappa, &
 &                      this%quark_mass(3)%kappa, &
 &                      this%smear_sous(1),       &
 &                      this%smear_sous(2),       &
 &                      this%smear_sous(3)
    do iop=0,NUM_BARYON_OP-1
      write(this%iout,'("op. ",I3," begin")')iop
      do itt=1,NTT
        write(this%iout,'(I3,2X,E15.8,2X,E15.8)')itt-1,this%prop(itt,iop)
      enddo
      write(this%iout,'("op. ",I3," end")')iop
    enddo
  endif

  return
end subroutine

#include "meson_localop.h90"
#include "meson_momentum.h90"
#include "meson_nonlocalop.h90"
#include "set_momentum_table.h90"
#include "baryon.h90"

end module
