#include <config.h>
module fftw3_class
  use comlib
  use lattice_class
  use print_status_class
  use timer_class
  use quark_prop_class
  use wavefunc_class
  implicit none
  private

  public :: new,delete
  public ::   fft_qprop
  public ::invfft_qprop
  public ::   fft_wavefunc
  public ::invfft_wavefunc
  public :: convolution

  integer, parameter :: NDIMT=1
  integer, parameter :: NVOL = (NTZ*NTY*NTX)


#include <fftw3.f>
  external dfftw_plan_dft_3d
  external dfftw_execute
  external dfftw_destroy_plan

  integer, parameter :: NSPACE=NX*NY*NZ

  logical, save :: m_is_initialized = .FALSE.

  type qprop_timecolspin_obj
    integer :: itt
    integer :: ic,jc
    integer :: is,js
    integer :: iqp,iqp_nl
    integer :: idummy
  end type

  integer, parameter :: NUM_QPROP = NTT*COL*COL*SPIN*SPIN
  type(qprop_timecolspin_obj), save :: qp_tsc(NUM_QPROP)                  ! GLOBAL table
  type(qprop_timecolspin_obj), save :: qp_tsc_nl(NUM_QPROP/NPU+1,0:NPU-1) ! node local table
  integer, save :: NUM_QPROP_NL   ! node local wavefunc number

  integer, save :: node_site(4,0:NPU-1)

contains

subroutine new_fft()
  implicit none
  integer :: iqp,iqp_nl
  integer :: iwf,iwf_nl
  integer :: itt,ic,is,jc,js
  integer :: ipu
  integer :: ipx, ipy, ipz, ipp
  integer :: ipux,ipuy,ipuz
  character(len=CHARLEN) :: str

  if (m_is_initialized) return

  node_site(:,:)=0
  node_site(1,nodeid)=ipsite(1)
  node_site(2,nodeid)=ipsite(2)
  node_site(3,nodeid)=ipsite(3)
  node_site(4,nodeid)= 0

  do ipu=0,NPU-1
    call comlib_bcast(node_site(:,ipu),ipu)
  enddo

!============================================
! make (time,col,col,spin,spin) 
!                 <=> qprop index iqp table
!============================================
  iqp=0
  do js=1,SPIN
  do is=1,SPIN
  do jc=1,COL
  do ic=1,COL
  do itt=1,NTT
    iqp=iqp+1
    qp_tsc(iqp)%itt=itt
    qp_tsc(iqp)%is =is
    qp_tsc(iqp)%js =js
    qp_tsc(iqp)%ic =ic
    qp_tsc(iqp)%jc =jc
    qp_tsc(iqp)%iqp=iqp
    qp_tsc(iqp)%iqp_nl=0 ! not used
    qp_tsc(iqp)%idummy=0 ! not used
  enddo
  enddo
  enddo
  enddo
  enddo

!============================================
! pick up qprop index for this MPI process
!============================================
  do ipu=0,NPU-1
  iqp_nl=0
  NUM_QPROP_NL=iqp_nl
  do iqp=1,NUM_QPROP
    if (mod(iqp-1,NPU) == ipu) then
      iqp_nl=iqp_nl+1
      NUM_QPROP_NL=iqp_nl
      qp_tsc_nl(iqp_nl,ipu) = qp_tsc(iqp)
      qp_tsc_nl(iqp_nl,ipu)%iqp_nl = iqp_nl
    endif
  enddo
  enddo

  m_is_initialized = .TRUE.

  return
end subroutine

subroutine fft_qprop(q)
!
! fft quark propagator
!
  implicit none
  complex(DP), intent(inout) :: q(COL,COL,SPIN,SPIN,NSPACE,NT)
  complex(DP), allocatable :: wf(:,:,:,:)
  integer :: iqp
  type(timer) :: timer1,timer2,timer3
  real(DP) :: etime1,etime2,etime3

  call new(timer1)
  call new(timer2)
  call new(timer3)

  call tic(timer1)

  if (.not.m_is_initialized) call new_fft()

  allocate(wf(NTZ,NTY,NTX,NUM_QPROP_NL))

  call tic(timer2)
  call gather_qprop(q,wf)
  call toc(timer2)

  call    fft_qprop_internal(wf,NUM_QPROP_NL)

  call tic(timer3)
  call scatter_qprop(wf,q)
  call toc(timer3)

  deallocate(wf)

  call toc(timer1)

  etime1 = get_elapse(timer1)
  etime2 = get_elapse(timer2)
  etime3 = get_elapse(timer3)
  if (nodeid==0) &
 &  write(*,'("%FFT_ETIME_QPROP: ALL=",F15.6," SCATTER_QPROP=",F15.6," GATHER_QPROP=",F15.6)') &
 &        etime1,etime2,etime3
  return
end subroutine

subroutine invfft_qprop(q)
!
! inverse fft quark propagator
!
  implicit none
  complex(DP), intent(inout) :: q(COL,COL,SPIN,SPIN,NSPACE,NT)
  complex(DP), allocatable :: wf(:,:,:,:)
  integer :: iqp
  type(timer) :: timer1,timer2,timer3
  real(DP) :: etime1,etime2,etime3
  call new(timer1)
  call new(timer2)
  call new(timer3)

  call tic(timer1)

  if (.not.m_is_initialized) call new_fft()

  allocate(wf(NTZ,NTY,NTX,NUM_QPROP_NL))

  call tic(timer2)
  call gather_qprop(q,wf)
  call toc(timer2)

  call invfft_qprop_internal(wf,NUM_QPROP_NL)

  call tic(timer3)
  call scatter_qprop(wf,q)
  call toc(timer3)

  deallocate(wf)
  call toc(timer1)

  etime1 = get_elapse(timer1)
  etime2 = get_elapse(timer2)
  etime3 = get_elapse(timer3)
  if (nodeid==0) &
 &  write(*,'("%INVFFT_ETIME_QPROP: ALL=",F15.6," SCATTER_QPROP=",F15.6," GATHER_QPROP=",F15.6)') &
 &        etime1,etime2,etime3

  return
end subroutine

subroutine fft_wavefunc(wf)
!
! fft sink wave function
!
  implicit none
  type(snk_wavefunc_obj), intent(inout) :: wf
  complex(DP), allocatable :: twf(:,:,:)
  type(timer) :: fft_time
  real(DP) :: etime

  call new(fft_time)

  call tic(fft_time)

  if (.not.m_is_initialized) call new_fft()

  allocate(twf(NTZ,NTY,NTX))
  call gather_wavefunc(wf,twf)
  call fft_wavefunc_internal(twf)
  call scatter_wavefunc(twf,wf)
  deallocate(twf)

  call toc(fft_time)

  etime = get_elapse(fft_time)
  if (nodeid==0)write(*,'("%FFT_ETIME_WAVEFUNC:",F15.6)')etime

  return
end subroutine

subroutine invfft_wavefunc(wf)
!
! inverse fft sink wave function
!
  implicit none
  type(snk_wavefunc_obj), intent(inout) :: wf
  complex(DP), allocatable :: twf(:,:,:)
  type(timer) :: fft_time
  real(DP) :: etime

  call new(fft_time)

  call tic(fft_time)

  if (.not.m_is_initialized) call new_fft()

  allocate(twf(NTZ,NTY,NTX))
  call gather_wavefunc(wf,twf)
  call invfft_wavefunc_internal(twf)
  call scatter_wavefunc(twf,wf)
  deallocate(twf)

  call toc(fft_time)

  etime = get_elapse(fft_time)
  if (nodeid==0)write(*,'("%INVFFT_ETIME_WAVEFUNC:",F15.6)')etime

  return
end subroutine

subroutine gather_qprop(q,wf)
!============================================================
! gather and redistribute quark propageotor for fft
!============================================================
  use mpi
  implicit none
  complex(DP), intent(in)  :: q(COL,COL,SPIN,SPIN,NSPACE,NT)
  complex(DP), intent(out) :: wf(NTZ,NTY,NTX,NUM_QPROP_NL)
  integer :: ispt,isp,ipu
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,jc,js,iqp_nl
  integer :: ipux,ipuy,ipuz,iput
  integer :: jpu, jqp_nl
  integer :: ierr
  complex(DP), allocatable :: tf(:,:),gf(:),ff(:,:,:,:)

  allocate(tf(NX*NY*NZ*NT,0:NPU-1))
  allocate(gf(NX*NY*NZ*NT))
  allocate(ff(NTZ,NTY,NTX,NTT))

  do js=1,SPIN
  do is=1,SPIN
  do jc=1,COL
  do ic=1,COL

!$OMP PARALLEL
!$OMP WORKSHARE
    tf(:,:) = Z0
!$OMP END WORKSHARE
!$OMP DO PRIVATE(it,ix,iy,iz,isp,ispt)
    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      isp = iz + (iy-1)*NZ + (ix-1)*NZ*NY
      ispt= isp + (it-1)*NZ*NY*NX
      gf(ispt) = q(ic,jc,is,js,isp,it)
    enddo
    enddo
    enddo
    enddo
!$OMP END DO
!$OMP END PARALLEL

    call comlib_allgather(gf,tf)

    do iput=0,NDIMT-1
    do ipux=0,NDIMX-1
    do ipuy=0,NDIMY-1
    do ipuz=0,NDIMZ-1
    do ipu=0,NPU-1
      if ( (node_site(1,ipu) == ipux) .and. &
  &        (node_site(2,ipu) == ipuy) .and. &
  &        (node_site(3,ipu) == ipuz) .and. &
  &        (node_site(4,ipu) == iput) ) then

!$OMP PARALLEL DO PRIVATE(it,ix,iy,iz,itt,itx,ity,itz,ispt)
        do it=1,NT
        do ix=1,NX
        do iy=1,NY
        do iz=1,NZ
          itt = it + NT*iput
          itx = ix + NX*ipux
          ity = iy + NY*ipuy
          itz = iz + NZ*ipuz
          ispt = iz + (iy-1)*NZ + (ix-1)*NZ*NY + (it-1)*NZ*NY*NX
          ff(itz,ity,itx,itt)=tf(ispt,ipu)
        enddo
        enddo
        enddo
        enddo
      endif
    enddo
    enddo
    enddo
    enddo
    enddo

    do itt=1,NTT
      do iqp_nl=1,NUM_QPROP_NL
        if ( (itt== qp_tsc_nl(iqp_nl,nodeid)%itt) .and. &
  &          (ic == qp_tsc_nl(iqp_nl,nodeid)%ic)  .and. &
  &          (jc == qp_tsc_nl(iqp_nl,nodeid)%jc)  .and. &
  &          (is == qp_tsc_nl(iqp_nl,nodeid)%is)  .and. &
  &          (js == qp_tsc_nl(iqp_nl,nodeid)%js) ) then
  
!$OMP PARALLEL DO PRIVATE(itx,ity,itz)
          do itx=1,NTX
          do ity=1,NTY
          do itz=1,NTZ
            wf(itz,ity,itx,iqp_nl)=ff(itz,ity,itx,itt)
          enddo
          enddo
          enddo
        endif
      enddo
    enddo

  enddo
  enddo
  enddo
  enddo

  deallocate(gf,tf,ff)
  return
end subroutine

subroutine scatter_qprop(wf,q)
!
! scatter quark propageotor for fft
!
  use mpi
  implicit none
  complex(DP), intent(in)    :: wf(NTZ,NTY,NTX,NUM_QPROP_NL)
  complex(DP), intent(inout) :: q(COL,COL,SPIN,SPIN,NSPACE,NT)
  integer :: ispt,isp,ipu,itsp
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,jc,js,iqp_nl
  integer :: ipux,ipuy,ipuz,iput
  integer :: ierr
  complex(DP), allocatable :: tf(:,:),gf(:)

!$OMP PARALLEL WORKSHARE
  q(:,:,:,:,:,:) = Z0
!$OMP END PARALLEL WORKSHARE

  allocate(tf(NTX*NTY*NTZ,0:NPU-1))
  allocate(gf(NTX*NTY*NTZ))
  do iqp_nl=1,NUM_QPROP_NL



!$OMP PARALLEL
!$OMP WORKSHARE
    tf(:,:) = Z0
!$OMP END WORKSHARE
!$OMP DO PRIVATE(itx,ity,itz,itsp)
    do itx=1,NTX
    do ity=1,NTY
    do itz=1,NTZ
      itsp = itz + (ity-1)*NTZ + (itx-1)*NTZ*NTY
      gf(itsp) = wf(itz,ity,itx,iqp_nl)
    enddo
    enddo
    enddo
!$OMP END DO
!$OMP END PARALLEL

    call comlib_allgather(gf,tf)

    do ipu=0,NPU-1
      itt = qp_tsc_nl(iqp_nl,ipu)%itt
      ic  = qp_tsc_nl(iqp_nl,ipu)%ic
      jc  = qp_tsc_nl(iqp_nl,ipu)%jc
      is  = qp_tsc_nl(iqp_nl,ipu)%is
      js  = qp_tsc_nl(iqp_nl,ipu)%js
!$OMP PARALLEL DO PRIVATE(itx,ity,itz,ix,iy,iz,it,ipux,ipuy,ipuz,iput,isp,itsp)
      do itx=1,NTX
      do ity=1,NTY
      do itz=1,NTZ
        ix = mod(itx-1,NX)+1
        iy = mod(ity-1,NY)+1
        iz = mod(itz-1,NZ)+1
        it = mod(itt-1,NT)+1
        ipux = (itx-1)/NX
        ipuy = (ity-1)/NY
        ipuz = (itz-1)/NZ
        iput = (itt-1)/NT
        if ( (ipux == ipsite(1)) .and.  &
  &          (ipuy == ipsite(2)) .and.  &
  &          (ipuz == ipsite(3)) .and.  &
  &          (iput == 0        )) then
          isp  = iz  + (iy -1)*NZ  + (ix -1)*NZ*NY
          itsp = itz + (ity-1)*NTZ + (itx-1)*NTZ*NTY
          q(ic,jc,is,js,isp,it) = tf(itsp,ipu)
        endif
      enddo
      enddo
      enddo
    enddo
  enddo

  deallocate(gf,tf)

  return
end subroutine

subroutine gather_wavefunc(wf,twf)
!
! gather sink wavefunction for fft
!
  implicit none
  type(snk_wavefunc_obj), intent(in) :: wf
  complex(DP), intent(out) :: twf(NTZ,NTY,NTX)
  integer :: ispt,isp,ipu
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,iwf_nl
  integer :: ipux,ipuy,ipuz,iput
  complex(DP), allocatable :: tf(:,:)
  complex(DP), allocatable :: gf(:)

  allocate(gf(NSPACE))
  allocate(tf(NSPACE,0:NPU-1))

!$OMP PARALLEL
!$OMP WORKSHARE
  tf(:,:) = Z0
!$OMP END WORKSHARE
!$OMP DO PRIVATE(isp)
  do isp=1,NSPACE
    gf(isp) = wf%wavefunc(isp)
  enddo
!$OMP END DO
!$OMP END PARALLEL

  call comlib_allgather(gf,tf)

!$OMP PARALLEL DO PRIVATE(itx,ity,itz,ix,iy,iz,ipux,ipuy,ipuz,isp,ipu)
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
    ix = mod(itx-1,NX)+1
    iy = mod(ity-1,NY)+1
    iz = mod(itz-1,NZ)+1
    ipux = (itx-1)/NX
    ipuy = (ity-1)/NY
    ipuz = (itz-1)/NZ
   
!    isp = iz + (iy-1)*NZ + (ix-1)*NZ*NY
    isp = ispace(ix,iy,iz)

    do ipu=0,NPU-1
    if ( (node_site(1,ipu) == ipux) .and. &
  &      (node_site(2,ipu) == ipuy) .and. &
  &      (node_site(3,ipu) == ipuz) .and. &
  &      (node_site(4,ipu) == 0   ) ) then 
      twf(itz,ity,itx)=tf(isp,ipu)
    endif
    enddo
  enddo
  enddo
  enddo

  deallocate(gf,tf)
#ifdef _DDD_
!===
if (0==nodeid) then
do itx=1,NTX
do ity=1,NTX
do itz=1,NTX
  write(*,'(3I3,2E24.15," FFTWF")')itx,ity,itz,twf(itz,ity,itx)
enddo
enddo
enddo
endif
!===
#endif

  return
end subroutine


subroutine scatter_wavefunc(twf,wf)
!
! scatter sink wave function for fft
!
  implicit none
  complex(DP), intent(in) :: twf(NTZ,NTY,NTX)
  type(snk_wavefunc_obj), intent(inout) :: wf
  integer :: ispt,isp,ipu,itsp
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,iwf_nl
  integer :: ipux,ipuy,ipuz,iput

!$OMP PARALLEL
!$OMP WORKSHARE
  wf%wavefunc(:) = Z0
!$OMP END WORKSHARE
!$OMP DO PRIVATE(itx,ity,itz,ix,iy,iz,ipux,ipuy,ipuz,isp)
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
    ix = mod(itx-1,NX)+1
    iy = mod(ity-1,NY)+1
    iz = mod(itz-1,NZ)+1
    ipux = (itx-1)/NX
    ipuy = (ity-1)/NY
    ipuz = (itz-1)/NZ
!    isp  = iz  + (iy -1)*NZ  + (ix -1)*NZ*NY
    isp  = ispace(ix,iy,iz)
    if ( (ipux == ipsite(1)) .and.  &
  &      (ipuy == ipsite(2)) .and.  &
  &      (ipuz == ipsite(3)) ) then 
       wf%wavefunc(isp) = twf(itz,ity,itx)
    endif
  enddo
  enddo
  enddo
!$OMP END DO
!$OMP END PARALLEL

  return
end subroutine

subroutine fft_qprop_internal(qp,nqp)
  implicit none
  integer,     intent(in)    :: nqp
  complex(DP), intent(inout) :: qp(NTZ,NTY,NTX,nqp)
  integer(DP) :: plan
  integer :: rank,howmany,istride,ostride,idist,odist
  integer :: rdim(3),inembed(3),onembed(3)
  rank=3
  howmany=nqp
  rdim(1)=NTZ
  rdim(2)=NTY
  rdim(3)=NTX
  inembed(:)=rdim(:)
  onembed(:)=rdim(:)
  istride=1
  ostride=1
  idist=NTZ*NTY*NTX
  odist=NTZ*NTY*NTX
  call dfftw_plan_many_dft(plan,rank,rdim,howmany,   &
 &                         qp,inembed,istride,idist, &
 &                         qp,onembed,ostride,odist, &
 &                         FFTW_FORWARD,FFTW_ESTIMATE)
  call dfftw_execute_dft(plan,qp,qp)
  call dfftw_destroy_plan(plan)
  return
end subroutine

subroutine invfft_qprop_internal(qp,nqp)
  implicit none
  integer,     intent(in)    :: nqp
  complex(DP), intent(inout) :: qp(NTZ,NTY,NTX,nqp)
  integer(DP) :: plan
  integer :: rank,howmany,istride,ostride,idist,odist
  integer :: rdim(3),inembed(3),onembed(3)
  real(DP) :: rfac
  rank=3
  howmany=nqp
  rdim(1)=NTZ
  rdim(2)=NTY
  rdim(3)=NTX
  inembed(:)=rdim(:)
  onembed(:)=rdim(:)
  istride=1
  ostride=1
  idist=NTZ*NTY*NTX
  odist=NTZ*NTY*NTX
  call dfftw_plan_many_dft(plan,rank,rdim,howmany,   &
 &                         qp,inembed,istride,idist, &
 &                         qp,onembed,ostride,odist, &
 &                         FFTW_BACKWARD,FFTW_ESTIMATE)
  call dfftw_execute_dft(plan,qp,qp)
  call dfftw_destroy_plan(plan)
  rfac = 1.0_DP/NVOL
  qp(:,:,:,:) = qp(:,:,:,:)*rfac
  return
end subroutine

subroutine fft_wavefunc_internal(af)
  implicit none
  complex(DP), intent(inout) :: af(NTZ,NTY,NTX)
  integer(DP) :: plan
  call dfftw_plan_dft_3d(plan,NTZ,NTY,NTX,af,af,FFTW_FORWARD,FFTW_ESTIMATE)
  call dfftw_execute_dft(plan,af,af)
  call dfftw_destroy_plan(plan)
  return
end subroutine

subroutine invfft_wavefunc_internal(af)
  implicit none
  complex(DP), intent(inout) :: af(NTZ,NTY,NTX)
  integer(DP) :: plan
  real(DP) :: rfac
  integer :: itx,ity,itz
  call dfftw_plan_dft_3d(plan,NTZ,NTY,NTX,af,af,FFTW_BACKWARD,FFTW_ESTIMATE)
  call dfftw_execute_dft(plan,af,af)
  call dfftw_destroy_plan(plan)
  rfac = 1.0_DP/NVOL
!$OMP PARALLEL DO PRIVATE(itx,ity,itz)
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
    af(itz,ity,itx) = af(itz,ity,itx)*rfac
  enddo
  enddo
  enddo
  return
end subroutine

subroutine convolution(wf,qp)
  implicit none
  complex(DP), intent(in)    :: wf(NSPACE)
  complex(DP), intent(inout) :: qp(COL,COL,SPIN,SPIN,NSPACE,NT)
  complex(DP) :: ctmp
  integer :: ic,jc,is,js,isp,it

  do it=1,NT
!$OMP PARALLEL DO PRIVATE(isp,ctmp,js,is,jc,ic)
    do isp=1,NSPACE
      ctmp = wf(isp)
      do js=1,SPIN
      do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        qp(ic,jc,is,js,isp,it) = qp(ic,jc,is,js,isp,it)*ctmp
      enddo
      enddo
      enddo
      enddo
    enddo
  enddo
 
  return
end subroutine

end module
