#$Id: make.inc,v 1.3 2011/09/30 09:54:34 ishikawa Exp $
#########################################################################
### include file for Makefile
#########################################################################
DATE = '"$(shell env LC_ALL=C date +"%b %d %Y")"'
TIME = '"$(shell env LC_ALL=C date +"%H:%M:%S")"'

RM	= /bin/rm -f
MV	= /bin/mv

#MACHINE = FUJITSU
MACHINE = INTEL
#MACHINE = GCC

###################################
# Intel Fortran
###################################
ifeq ($(MACHINE),INTEL)

MPIHOME=/opt/openmpi-1.4.3

#PAPIDIR = /opt/papi-4.1.2.1
#PAPILIB= -L$(PAPIDIR)/lib -lpapi
#PAPIINC= -I$(PAPIDIR)/include
USEGPU = -D_GPU_SOLVER_
CUDADIR = /opt/cuda
CUDASDKDIR = /opt/NVIDIA_GPU_Computing_SDK/C
CUDALIBS = ../GPUSolverClass/libgpusolver.a \
           ../GPUBlockSolverClass/libgpublksolver.a \
           -L$(CUDADIR)/lib64 -L$(CUDASDKDIR)/lib -lcudart -lcutil_x86_64 -lm 

#MKLPATH=$(MKLROOT)/lib/em64t
MKLPATH=$(MKLROOT)/lib/intel64
MKLINC = -I$(MKLROOT)/include
MKLLIB = -L$(MKLPATH) \
          -Wl,--start-group $(MKLPATH)/libmkl_intel_lp64.a \
                            $(MKLPATH)/libmkl_intel_thread.a \
                            $(MKLPATH)/libmkl_core.a \
          -Wl,--end-group -liomp5 -lpthread

LAPACK95INC= -I$(HOME)/include/lapack95_modules
LAPACK95LIB= -L$(HOME)/lib -llapack95

USEQMX = -D_QMX_
OMXPATH=/opt/open-mx-1.5.0
QMXINC = -I../Qmxlib -I$(OMXPATH)/include
QMXLIB = ../Qmxlib/qmx_lib.o ../Qmxlib/qmxf03.o -lrt $(OMXPATH)/lib/libmyriexpress.a -lstdc++

AR      = xiar
RANLIB  = ranlib

FC      = $(MPIHOME)/bin/mpif90
#CC      = icc -O2 -align $(OPENMP)
CC      = icc -O2 -align
CPP     = cpp

OPENMP  = -openmp -openmp-report2
#OPENMP  = 

#DEBUG   = -p -g -traceback -D_DEBUG
#DEBUG   = -g -traceback -D_DEBUG -D_PERFMON_
#DEBUG   = -g -traceback -D_DEBUG
#DEBUG   = -g -traceback
DEBUG   = -check all -debug all -g -traceback
#DEBUG   = -g -traceback -vec-report3 -opt-report
#DEBUG   = -check all -debug all -g -traceback -fpe-all=0 -fp-stack-check -D_DEBUG
#DEBUG   =  -fpe-all=0 -fp-stack-check -check all -debug all -g -traceback
#DEBUG   = -check all -g -traceback -fpe-all=0 -fp-stack-check -D_DEBUG
#DEBUG   = -g -traceback -fpe-all=0 -D_DEBUG
#DEBUG   = -p -g -traceback
#DEBUG   = -g -traceback -warn all
#FFLAGS  = -O2 -xSSE4.2 -stand f03 -i4 -u -static-intel -align all $(DEBUG) $(OPENMP) $(USEGPU)
FFLAGS  = -O0 -xSSE4.2 -stand f03 -i4 -u -shared-intel -mcmodel=medium -align all $(DEBUG) $(OPENMP) $(USEGPU)

#FFLAGS  = -O2 -xSSE4.2 -stand f03 -i4 -u -static-intel -mcmodel=medium -align all \
#          -finline -finline-limit=100000 -inline-factor=10000 -override-limits \
#          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)

LDFLAGS = $(FFLAGS)
MOD     = mod

LIBS    = $(LAPACK95LIB) $(MKLLIB) $(PAPILIB) $(QMXLIB) $(CUDALIBS)
INCLUDE = $(MKLINC) $(LAPACK95INC) $(PAPIINC) $(QMXINC)

endif

###################################
# GFortran 4.6.1
###################################
ifeq ($(MACHINE),GCC)

MPIHOME=/opt/openmpi-1.4.3-gcc

USEGPU = -D_GPU_SOLVER_
CUDADIR = /opt/cuda
CUDASDKDIR = /opt/NVIDIA_GPU_Computing_SDK/C
CUDALIBS = ../GPUSolverClass/libgpusolver.a \
           ../GPUBlockSolverClass/libgpublksolver.a \
           -L$(CUDADIR)/lib64 -L$(CUDASDKDIR)/lib -lcudart -lcutil_x86_64 -lm 

LAPACK95INC= -I$(HOME)/include/lapack95_modules_gfortran4.6.0
LAPACK95LIB= -L$(HOME)/lib -llapack95_gfortran4.6.0  -llapack_gcc4.6.0 -lblas_gcc4.6.0

USEQMX =
#USEQMX = -D_QMX_
#OMXPATH=/opt/open-mx-1.4.0
#QMXINC = -I../Qmxlib -I$(OMXPATH)/include
#QMXLIB = ../Qmxlib/qmx_lib.o ../Qmxlib/qmxf03.o -lrt -L$(OMXPATH)/lib -lmyriexpress -lstdc++

AR      = ar
RANLIB  = ranlib
FC      = $(MPIHOME)/bin/mpif90                  
CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
#FC      = $(MPIHOME)/bin/mpif90                  
#CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
#FC      = $(MPIHOME)/bin/mpif90                  
#CC      = $(MPIHOME)/bin/mpicc -O2 -falign-labels
CPP     = gcc
OPENMP  = -fopenmp
#OPENMP  = 
DEBUG   = -g -fbacktrace
#DEBUG   = -g -fbacktrace -D_DEBUG
#DEBUG   = -g -fbacktrace -frange-check -fbounds-check -fvar-tracking -D_DEBUG
#FFLAGS  = -O3 -msse4.2 -std=f2003 -finline-functions -falign-labels -fall-intrinsics \
#          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)
FFLAGS  = -O0 -msse4.2  -finline-functions -falign-labels -fall-intrinsics \
          $(DEBUG) $(OPENMP) $(USEQMX) $(USEGPU)
LDFLAGS = $(FFLAGS)
MOD     = mod

LIBS    = $(LAPACK95LIB) $(MKLLIB) $(PAPILIB) $(QMXLIB) $(CUDALIBS)
INCLUDE = $(MKLINC) $(LAPACK95INC) $(PAPIINC) $(QMXINC)

endif

###################################
# Fujitsu Fortran
###################################
ifeq ($(MACHINE),FUJITSU)
AR      = ar
FC      = mpif90 -nochoicemod
CC      = cc
CPP     = cpp
#DEBUG   = -E cgitu -H aesux -Qa,d,i,x
DEBUG   =
FFLAGS  = -static -w -X9 -Am -v95e -K fast,fastlib,SSE2,PENTIUM4,prefetch=2 \
          $(DEBUG)
LDFLAGS = $(FFLAGS)
#FFTLIBS = -lssl2
LIBS    = 
MOD     = mod
endif


###################################
INCLUDE := $(INCLUDE) $(shell ls -1 -d ../*Class ../*Module | awk '{print "-I"$$1" "}')


OBJS    = $(SOURCES:.F90=.o)
LOGS    = $(SOURCES:.F90=.diag) $(SOURCES:.F90=.L) \
          $(SOURCES:.F90=.par)  $(SOURCES:.F90=.log) \
          $(SOURCES:.F90=.lst)

.F90.o :

%.o : %.F90
	$(FC) $(FFLAGS) $(MONITOR) $(INCLUDE) -c $< -o $@

%.s : %.F90
	$(FC) $(FFLAGS) $(MONITOR) $(INCLUDE) -S $< -o $@

#########################################################################
